#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

from pyomo.environ import (
    Var,
    Param,
    Set,
    Suffix,
    NonNegativeReals,
    PositiveIntegers,
    units as pyunits,
    check_optimal_termination,
)
from pyomo.common.config import Bool, ConfigBlock, ConfigValue, In
from enum import Enum, auto

from idaes.core import (
    declare_process_block_class,
    MaterialBalanceType,
    EnergyBalanceType,
    MomentumBalanceType,
    UnitModelBlockData,
    useDefault,
)
from idaes.core.solvers import get_solver
from idaes.core.util.constants import Constants
from idaes.core.util.config import is_physical_parameter_block
from idaes.core.util.tables import create_stream_table_dataframe
from idaes.core.util.exceptions import ConfigurationError, InitializationError
import idaes.core.util.scaling as iscale
import idaes.logger as idaeslog

from watertap.core import ControlVolume0DBlock, InitializationMixin
from watertap.costing.unit_models.gac import cost_gac

__author__ = "Hunter Barber"

_log = idaeslog.getLogger(__name__)


# ---------------------------------------------------------------------
class FilmTransferCoefficientType(Enum):
    fixed = auto()  # liquid phase film transfer coefficient is a user specified value
    calculated = (
        auto()
    )  # calculate liquid phase film transfer coefficient from the Gnielinshi correlation


class SurfaceDiffusionCoefficientType(Enum):
    fixed = auto()  # surface diffusion coefficient is a user specified value
    calculated = auto()  # calculate surface diffusion coefficient


# ---------------------------------------------------------------------
@declare_process_block_class("GAC")
class GACData(InitializationMixin, UnitModelBlockData):
    """
    Empirical Constant-Pattern-Homogeneous-Surface-Diffusion Model (CPHSDM) for Granular Activated Carbon -
    currently only available to model with the multicomp_aq_sol_prop_pack
    """

    CONFIG = ConfigBlock()

    CONFIG.declare(
        "dynamic",
        ConfigValue(
            domain=In([False]),
            default=False,
            description="Dynamic model flag - must be False",
            doc="""Indicates whether this model will be dynamic or not,
    **default** = False. The filtration unit does not support dynamic
    behavior, thus this must be False.""",
        ),
    )
    CONFIG.declare(
        "has_holdup",
        ConfigValue(
            default=False,
            domain=In([False]),
            description="Holdup construction flag - must be False",
            doc="""Indicates whether holdup terms should be constructed or not.
    **default** - False. The filtration unit does not have defined volume, thus
    this must be False.""",
        ),
    )
    CONFIG.declare(
        "property_package",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="Property package to use for control volume",
            doc="""Property parameter object used to define property calculations,
    **default** - useDefault.
    **Valid values:** {
    **useDefault** - use default package from parent model or flowsheet,
    **PhysicalParameterObject** - a PhysicalParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing property packages",
            doc="""A ConfigBlock with arguments to be passed to a property block(s)
    and used when constructing these,
    **default** - None.
    **Valid values:** {
    see property package for documentation.}""",
        ),
    )
    CONFIG.declare(
        "material_balance_type",
        ConfigValue(
            default=MaterialBalanceType.useDefault,
            domain=In(MaterialBalanceType),
            description="Material balance construction flag",
            doc="""Indicates what type of mass balance should be constructed,
    **default** - MaterialBalanceType.useDefault.
    **Valid values:** {
    **MaterialBalanceType.useDefault - refer to property package for default
    balance type
    **MaterialBalanceType.none** - exclude material balances,
    **MaterialBalanceType.componentPhase** - use phase component balances,
    **MaterialBalanceType.componentTotal** - use total component balances,
    **MaterialBalanceType.elementTotal** - use total element balances,
    **MaterialBalanceType.total** - use total material balance.}""",
        ),
    )
    CONFIG.declare(
        "is_isothermal",
        ConfigValue(
            default=True,
            domain=Bool,
            description="""Assume isothermal conditions for control volume(s); energy_balance_type must be EnergyBalanceType.none,
    **default** - True.""",
        ),
    )

    CONFIG.declare(
        "energy_balance_type",
        ConfigValue(
            default=EnergyBalanceType.none,
            domain=In(EnergyBalanceType),
            description="Energy balance construction flag",
            doc="""Indicates what type of energy balance should be constructed,
    **default** - EnergyBalanceType.none.
    **Valid values:** {
    **EnergyBalanceType.useDefault - refer to property package for default
    balance type
    **EnergyBalanceType.none** - exclude energy balances,
    **EnergyBalanceType.enthalpyTotal** - single enthalpy balance for material,
    **EnergyBalanceType.enthalpyPhase** - enthalpy balances for each phase,
    **EnergyBalanceType.energyTotal** - single energy balance for material,
    **EnergyBalanceType.energyPhase** - energy balances for each phase.}""",
        ),
    )
    CONFIG.declare(
        "momentum_balance_type",
        ConfigValue(
            default=MomentumBalanceType.pressureTotal,
            domain=In(MomentumBalanceType),
            description="Momentum balance construction flag",
            doc="""Indicates what type of momentum balance should be constructed,
        **default** - MomentumBalanceType.pressureTotal.
        **Valid values:** {
        **MomentumBalanceType.none** - exclude momentum balances,
        **MomentumBalanceType.pressureTotal** - single pressure balance for material,
        **MomentumBalanceType.pressurePhase** - pressure balances for each phase,
        **MomentumBalanceType.momentumTotal** - single momentum balance for material,
        **MomentumBalanceType.momentumPhase** - momentum balances for each phase.}""",
        ),
    )
    CONFIG.declare(
        "film_transfer_coefficient_type",
        ConfigValue(
            default=FilmTransferCoefficientType.fixed,
            domain=In(FilmTransferCoefficientType),
            description="Surface diffusion coefficient",
            doc="""Indicates whether the liquid phase film transfer rate will be calculated or fixed by the user
        **default** - FilmTransferCoefficientType.fixed.
        **Valid values:** {
        **FilmTransferCoefficientType.fixed** - user specifies film transfer rate,
        **FilmTransferCoefficientType.calculated** - calculates film transfer rate based on the Gnielinshi correlation}""",
        ),
    )
    CONFIG.declare(
        "surface_diffusion_coefficient_type",
        ConfigValue(
            default=SurfaceDiffusionCoefficientType.fixed,
            domain=In(SurfaceDiffusionCoefficientType),
            description="Surface diffusion coefficient",
            doc="""Indicates whether the surface diffusion coefficient will be calculated or fixed by the user
        **default** - SurfaceDiffusionCoefficientType.fixed.
        **Valid values:** {
        **SurfaceDiffusionCoefficientType.fixed** - user specifies surface diffusion coefficient,
        **SurfaceDiffusionCoefficientType.calculated** - calculates surface diffusion coefficient by the Crittenden correlation}""",
        ),
    )
    CONFIG.declare(
        "target_species",
        ConfigValue(
            default=None,
            domain=list,
            description="Species target for adsorption, currently only supports single species",
            doc="""Indicate which component in the property package's component list is the target species
        for adsorption by the GAC system, currently the model supports a single species
        **default** - None.
        **Valid values:** {
        if the property package solute set only contains one item (two component, one solute and one solvent/water),
        the model will accept the single solute as the target species, for multi-solute systems a string of
        the component id must be provided.}""",
        ),
    )
    CONFIG.declare(
        "finite_elements_ss_approximation",
        ConfigValue(
            default=5,
            domain=int,
            description="Number of finite elements in operational time domain for steady state approximation",
            doc="""Number of finite elements to use when discretizing operational time (default=5)""",
        ),
    )

    def _validate_config(self):
        if (
            self.config.is_isothermal
            and self.config.energy_balance_type != EnergyBalanceType.none
        ):
            raise ConfigurationError(
                "If the isothermal assumption is used then the energy balance type must be none"
            )

    # ---------------------------------------------------------------------
    def build(self):

        super().build()

        # create blank scaling factors to be populated later
        self.scaling_factor = Suffix(direction=Suffix.EXPORT)
        # get default units from property package
        units_meta = self.config.property_package.get_metadata().get_derived_units
        # Check configs for errors
        self._validate_config()

        # ---------------------------------------------------------------------
        # separate target_species to be adsorbed and other species considered inert
        # apply target_species automatically if arg left to default and only one viable option exists

        if self.config.target_species is None:
            if len(self.config.property_package.solute_set) == 1:
                self.config.target_species = self.config.property_package.solute_set
                self.target_species = self.config.target_species
            else:
                raise ConfigurationError(
                    "'target species' is not specified for the GAC unit model, either specify 'target species'"
                    " argument or reduce solute set to a single component"
                )
        else:
            self.target_species = Set(dimen=1)
            for str_species in self.config.target_species:
                if not isinstance(str_species, str):
                    raise ConfigurationError(
                        f"item {str_species} within 'target_species' list is not of data type str"
                    )
                if str_species not in self.config.property_package.component_list:
                    raise ConfigurationError(
                        f"item {str_species} within 'target_species' list is not in 'component_list'"
                    )
                self.target_species.add(str_species)

        self.inert_species = Set(
            dimen=1,
            initialize=(
                self.config.property_package.component_list - self.config.target_species
            ),
        )

        # ---------------------------------------------------------------------
        # build control volume
        self.process_flow = ControlVolume0DBlock(
            dynamic=False,
            has_holdup=False,
            property_package=self.config.property_package,
            property_package_args=self.config.property_package_args,
        )
        self.process_flow.add_state_blocks(has_phase_equilibrium=False)
        self.process_flow.add_material_balances(
            balance_type=self.config.material_balance_type, has_mass_transfer=True
        )
        self.process_flow.add_energy_balances(
            balance_type=self.config.energy_balance_type,
            has_enthalpy_transfer=False,
        )
        self.process_flow.add_momentum_balances(
            balance_type=self.config.momentum_balance_type,
            has_pressure_change=False,
        )
        if self.config.is_isothermal:
            self.process_flow.add_isothermal_assumption()

        # add port for adsorbed contaminant contained in nearly saturated GAC particles
        tmp_dict = dict(**self.config.property_package_args)
        tmp_dict["has_phase_equilibrium"] = False
        tmp_dict["parameters"] = self.config.property_package
        tmp_dict["defined_state"] = False
        self.gac_removed = self.config.property_package.state_block_class(
            self.flowsheet().config.time,
            doc="state block of the species contained within the gac removed from the bed",
            **tmp_dict,
        )

        # Add ports
        self.add_inlet_port(name="inlet", block=self.process_flow)
        self.add_outlet_port(name="outlet", block=self.process_flow)
        self.add_outlet_port(name="adsorbed", block=self.gac_removed)

        # ---------------------------------------------------------------------
        # Freundlich isotherm parameters and other property variables

        self.freund_k = Var(
            initialize=10,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=pyunits.dimensionless,  # dynamic with freund_ninv, ((length ** 3) * (mass ** -1)) ** freund_ninv,
            doc="Freundlich isotherm k parameter, must be provided in base [L3/M] units",
        )
        self.freund_ninv = Var(
            initialize=0.5,
            bounds=(0, 1),
            domain=NonNegativeReals,
            units=pyunits.dimensionless,
            doc="Freundlich isotherm 1/n parameter",
        )
        self.ds = Var(
            initialize=1e-14,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=units_meta("length") ** 2 * units_meta("time") ** -1,
            doc="surface diffusion coefficient",
        )
        self.kf = Var(
            initialize=1e-5,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=units_meta("length") * units_meta("time") ** -1,
            doc="liquid phase film transfer coefficient",
        )
        self.equil_conc = Var(
            initialize=1e-4,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=pyunits.dimensionless,
            doc="equilibrium concentration of adsorbed phase with liquid phase",
        )
        self.dg = Var(
            initialize=1e5,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=pyunits.dimensionless,
            doc="solute distribution parameter",
        )
        self.N_Bi = Var(
            initialize=10,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=pyunits.dimensionless,
            doc="Biot number",
        )

        # ---------------------------------------------------------------------
        # bed dimensions and related variables

        self.velocity_sup = Var(
            initialize=0.001,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=units_meta("length") * units_meta("time") ** -1,
            doc="superficial velocity",
        )
        self.velocity_int = Var(
            initialize=0.002,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=units_meta("length") * units_meta("time") ** -1,
            doc="interstitial velocity",
        )
        self.bed_voidage = Var(
            initialize=0.5,
            bounds=(0, 1),
            domain=NonNegativeReals,
            units=pyunits.dimensionless,
            doc="bed void fraction",
        )
        self.bed_length = Var(
            initialize=5,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=units_meta("length"),
            doc="bed length",
        )
        self.bed_diameter = Var(
            initialize=1,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=units_meta("length"),
            doc="bed diameter, valid if considering only a single adsorber",
        )
        self.bed_area = Var(
            initialize=1,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=units_meta("length") ** 2,
            doc="bed area, single adsorber area or sum of areas for adsorbers in parallel",
        )
        self.bed_volume = Var(
            initialize=5,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=units_meta("length") ** 3,
            doc="bed volume",
        )
        self.ebct = Var(
            initialize=500,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=units_meta("time"),
            doc="empty bed contact time",
        )
        self.residence_time = Var(
            initialize=100,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=units_meta("time"),
            doc="fluid residence time in the bed",
        )
        self.bed_mass_gac = Var(
            initialize=100,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=units_meta("mass"),
            doc="mass of fresh gac in the bed",
        )

        # ---------------------------------------------------------------------
        # gac particle properties
        # TODO: Potentially create a default option

        self.particle_dens_app = Var(
            initialize=1000,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=units_meta("mass") * units_meta("length") ** -3,
            doc="gac apparent density",
        )
        self.particle_dens_bulk = Var(
            initialize=500,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=units_meta("mass") * units_meta("length") ** -3,
            doc="gac bulk density",
        )
        self.particle_dia = Var(
            initialize=0.001,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=units_meta("length"),
            doc="gac particle diameter",
        )

        # ---------------------------------------------------------------------
        # constants in empirical equations

        self.a0 = Var(
            initialize=1,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=pyunits.dimensionless,
            doc="Stanton equation parameter 0",
        )
        self.a1 = Var(
            initialize=1,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=pyunits.dimensionless,
            doc="Stanton equation parameter 1",
        )
        self.b0 = Var(
            initialize=0.1,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=pyunits.dimensionless,
            doc="throughput equation parameter 0",
        )
        self.b1 = Var(
            initialize=0.1,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=pyunits.dimensionless,
            doc="throughput equation parameter 1",
        )
        self.b2 = Var(
            initialize=0.1,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=pyunits.dimensionless,
            doc="throughput equation parameter 2",
        )
        self.b3 = Var(
            initialize=0.1,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=pyunits.dimensionless,
            doc="throughput equation parameter 3",
        )
        self.b4 = Var(
            initialize=0.1,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=pyunits.dimensionless,
            doc="throughput equation parameter 4",
        )

        # ---------------------------------------------------------------------
        # conditions to achieve a constant pattern solution

        self.min_N_St = Var(
            initialize=10,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=pyunits.dimensionless,
            doc="minimum Stanton number to achieve a constant pattern solution",
        )
        self.min_ebct = Var(
            initialize=500,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=units_meta("time"),
            doc="minimum empty bed contact time to achieve a constant pattern solution",
        )
        self.throughput = Var(
            initialize=1,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=pyunits.dimensionless,
            doc="specific throughput from empirical equation",
        )
        self.min_residence_time = Var(
            initialize=1000,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=units_meta("time"),
            doc="minimum fluid residence time in the bed to achieve a constant pattern solution",
        )
        self.min_operational_time = Var(
            initialize=1e8,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=units_meta("time"),
            doc="minimum operational time of the bed from fresh to achieve a constant pattern solution",
        )

        # ---------------------------------------------------------------------
        # performance variables of CPHSDM

        self.conc_ratio_replace = Var(
            initialize=0.5,
            bounds=(0, 1),
            domain=NonNegativeReals,
            units=pyunits.dimensionless,
            doc="effluent to inlet concentration ratio at operational time",
        )
        self.operational_time = Var(
            initialize=1e5,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=units_meta("time"),
            doc="operational time of the bed from fresh",
        )
        self.bed_volumes_treated = Var(
            initialize=10000,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=pyunits.dimensionless,
            doc="bed volumes treated at operational time",
        )

        # ---------------------------------------------------------------------
        # steady state approximation

        self.elements_ss_approx = Param(
            mutable=True,
            default=5,
            initialize=self.config.finite_elements_ss_approximation,
            domain=PositiveIntegers,
            units=pyunits.dimensionless,
            doc="number of discretized operational time elements used for steady state approximation",
        )
        self.conc_ratio_start_breakthrough = Param(
            mutable=True,
            default=0.01,
            initialize=0.01,
            domain=NonNegativeReals,
            units=pyunits.dimensionless,
            doc="the concentration ratio at which the breakthrough curve is to start, typically between 0.01 amd 0.05",
        )

        # create index for discretized elements with element [0] containing 0s for conc ratio and operational time
        ele_disc = range(self.elements_ss_approx.value + 1)
        ele_index = ele_disc[1:]

        self.ele_throughput = Var(
            ele_index,
            initialize=1,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=pyunits.dimensionless,
            doc="specific throughput from empirical equation by discrete element",
        )
        self.ele_min_operational_time = Var(
            ele_index,
            initialize=1e8,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=units_meta("time"),
            doc="minimum operational time of the bed from fresh to achieve a constant pattern solution by discrete element",
        )
        self.ele_conc_ratio_replace = Var(
            ele_disc,
            initialize=0.05,
            bounds=(0, 1),
            domain=NonNegativeReals,
            units=pyunits.dimensionless,
            doc="effluent to inlet concentration ratio at operational time by discrete element",
        )
        self.ele_operational_time = Var(
            ele_disc,
            initialize=1e5,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=units_meta("time"),
            doc="operational time of the bed from fresh by discrete element",
        )
        self.ele_conc_ratio_term = Var(
            ele_index,
            initialize=0.05,
            bounds=(0, 1),
            domain=NonNegativeReals,
            units=pyunits.dimensionless,
            doc="trapezoid rule of elements for numerical integration of average concentration ratio",
        )
        self.conc_ratio_avg = Var(
            initialize=0.1,
            bounds=(0, 1),
            domain=NonNegativeReals,
            units=pyunits.dimensionless,
            doc="steady state approximation of average effluent to inlet concentration ratio in operational time by trapezoid rule",
        )
        self.mass_adsorbed = Var(
            initialize=10,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=units_meta("mass"),
            doc="total mass of adsorbed species at operational time",
        )
        self.gac_usage_rate = Var(
            initialize=1e-2,
            bounds=(0, None),
            domain=NonNegativeReals,
            units=units_meta("mass") * units_meta("time") ** -1,
            doc="gac usage/replacement/regeneration rate",
        )

        # ---------------------------------------------------------------------
        # property equations and other dimensionless variables

        @self.Constraint(
            self.flowsheet().config.time,
            self.target_species,
            doc="equilibrium concentration",
        )
        def eq_equilibrium_concentration(b, t, j):
            freund_k_units = (
                (units_meta("length") ** 3) * units_meta("mass") ** -1
            ) ** b.freund_ninv
            return b.equil_conc == b.freund_k * freund_k_units * (
                b.process_flow.properties_in[t].conc_mass_phase_comp["Liq", j]
                ** b.freund_ninv
            )

        @self.Constraint(
            self.flowsheet().config.time,
            self.target_species,
            doc="solute distribution parameter",
        )
        def eq_dg(b, t, j):
            return b.dg * b.bed_voidage * b.process_flow.properties_in[
                t
            ].conc_mass_phase_comp["Liq", j] == b.particle_dens_app * b.equil_conc * (
                1 - b.bed_voidage
            )

        @self.Constraint(doc="Biot number")
        def eq_number_bi(b):
            return b.N_Bi * b.ds * b.dg * b.bed_voidage == b.kf * (
                b.particle_dia / 2
            ) * (1 - b.bed_voidage)

        # ---------------------------------------------------------------------
        # bed dimensions, gac particle, and sizing calculations

        @self.Constraint(doc="bed void fraction based on gac particle densities")
        def eq_bed_voidage(b):
            return b.bed_voidage == 1 - (b.particle_dens_bulk / b.particle_dens_app)

        @self.Constraint(doc="relating velocities based on bed voidage")
        def eq_velocity_relation(b):
            return b.velocity_sup == b.velocity_int * b.bed_voidage

        @self.Constraint(doc="bed length based on velocity and ebct")
        def eq_bed_length(b):
            return b.bed_length == b.velocity_sup * b.ebct

        @self.Constraint(doc="bed diameter and area relation")
        def eq_bed_diameter(b):
            return Constants.pi * (b.bed_diameter**2) == 4 * b.bed_area

        @self.Constraint(
            self.flowsheet().config.time,
            doc="bed area based on velocity and volumetric flow",
        )
        def eq_bed_area(b, t):
            return (
                b.bed_area * b.velocity_sup
                == b.process_flow.properties_in[t].flow_vol_phase["Liq"]
            )

        @self.Constraint(doc="bed volume based on cylindrical dimensions")
        def eq_bed_volume(b):
            return b.bed_volume == b.bed_length * b.bed_area

        @self.Constraint(doc="fluid residence time in the bed")
        def eq_residence_time(b):
            return b.residence_time == b.bed_voidage * b.ebct

        @self.Constraint(doc="total mass of gac in the bed")
        def eq_mass_gac_bed(b):
            return b.bed_mass_gac == b.particle_dens_bulk * b.bed_volume

        # ---------------------------------------------------------------------
        # gac CPHSDM intermediate equations

        @self.Constraint(
            doc="minimum Stanton number to achieve constant pattern solution"
        )
        def eq_min_number_st_cps(b):
            return b.min_N_St == b.a0 * b.N_Bi + b.a1

        @self.Constraint(
            doc="minimum empty bed contact time to achieve constant pattern solution"
        )
        def eq_min_ebct_cps(b):
            return b.min_ebct * (1 - b.bed_voidage) * b.kf == b.min_N_St * (
                b.particle_dia / 2
            )

        @self.Constraint(doc="throughput based on empirical 5-parameter regression")
        def eq_throughput(b):
            return b.throughput == b.b0 + b.b1 * (
                b.conc_ratio_replace**b.b2
            ) + b.b3 / (1.01 - (b.conc_ratio_replace**b.b4))

        @self.Constraint(
            doc="minimum fluid residence time in the bed to achieve a constant pattern solution"
        )
        def eq_min_residence_time_cps(b):
            return b.min_residence_time == b.bed_voidage * b.min_ebct

        @self.Constraint(
            doc="minimum operational time of the bed from fresh to achieve a constant pattern solution"
        )
        def eq_minimum_operational_time_cps(b):
            return (
                b.min_operational_time
                == b.min_residence_time * (b.dg + 1) * b.throughput
            )

        # ---------------------------------------------------------------------
        # gac performance equations

        @self.Constraint(
            doc="elapsed operational time between a fresh bed and the theoretical bed replacement"
        )
        def eq_operational_time(b):
            return b.operational_time == b.min_operational_time + (
                b.residence_time - b.min_residence_time
            ) * (b.dg + 1)

        @self.Constraint(doc="bed volumes treated")
        def eq_bed_volumes_treated(b):
            return (
                b.bed_volumes_treated * b.residence_time
                == b.operational_time * b.bed_voidage
            )

        # ---------------------------------------------------------------------
        # steady state approximation

        self.ele_conc_ratio_replace[0].fix(0)
        self.ele_operational_time[0].fix(0)

        @self.Constraint(
            ele_index,
            doc="throughput based on empirical 5-parameter regression by discretized element",
        )
        def eq_ele_throughput(b, ele):
            return b.ele_throughput[ele] == b.b0 + b.b1 * (
                b.ele_conc_ratio_replace[ele] ** b.b2
            ) + b.b3 / (1.01 - b.ele_conc_ratio_replace[ele] ** b.b4)

        @self.Constraint(
            ele_index,
            doc="minimum operational time of the bed from fresh to achieve a constant pattern solution by discretized element",
        )
        def eq_ele_min_operational_time(b, ele):
            return (
                b.ele_min_operational_time[ele]
                == b.min_residence_time * (b.dg + 1) * b.ele_throughput[ele]
            )

        @self.Constraint(
            ele_index,
            doc="creating evenly spaced discretized elements",
        )
        def eq_ele_conc_ratio_replace(b, ele):
            return b.ele_conc_ratio_replace[ele] == b.conc_ratio_start_breakthrough + (
                ele_disc[ele] - 1
            ) * (
                (b.conc_ratio_replace - b.conc_ratio_start_breakthrough)
                / (b.elements_ss_approx - 1)
            )

        @self.Constraint(
            ele_index,
            doc="operational time of the bed by discretized element",
        )
        def eq_ele_operational_time(b, ele):
            return b.ele_operational_time[ele] == b.ele_min_operational_time[ele] + (
                b.residence_time - b.min_residence_time
            ) * (b.dg + 1)

        @self.Constraint(
            ele_index,
            doc="finite element discretization of concentration ratios over time",
        )
        def eq_ele_conc_ratio_term(b, ele):
            return b.ele_conc_ratio_term[ele] == (
                (b.ele_operational_time[ele] - b.ele_operational_time[ele - 1])
                / b.ele_operational_time[self.elements_ss_approx.value]
            ) * (
                (b.ele_conc_ratio_replace[ele] + b.ele_conc_ratio_replace[ele - 1]) / 2
            )

        @self.Constraint(
            doc="summation of finite elements for average concentration during operating time"
        )
        def eq_conc_ratio_avg(b):
            return b.conc_ratio_avg == sum(
                b.ele_conc_ratio_term[ele] for ele in ele_index
            )

        @self.Constraint(
            self.target_species,
            doc="mass adsorbed in the operational time",
        )
        def eq_mass_adsorbed(b, j):
            return (
                b.mass_adsorbed
                == b.gac_removed[0].flow_mass_phase_comp["Liq", j] * b.operational_time
            )

        @self.Constraint(doc="steady state rate of new gac mass required")
        def eq_gac_usage_rate(b):
            return b.gac_usage_rate * b.operational_time == b.bed_mass_gac

        # ---------------------------------------------------------------------
        # balance equations

        # isothermal for port not in control volume
        @self.Constraint(
            self.flowsheet().config.time,
            doc="isothermal assumption for the species contained within the gac removed from the bed",
        )
        def eq_isothermal_gac_removed(b, t):
            return (
                b.process_flow.properties_in[t].temperature
                == b.gac_removed[t].temperature
            )

        # isobaric for port not in control volume
        @self.Constraint(
            self.flowsheet().config.time,
            doc="isobaric assumption for the species contained within the gac removed from the bed",
        )
        def eq_isobaric_gac_removed(b, t):
            return b.process_flow.properties_in[t].pressure == b.gac_removed[t].pressure

        # mass transfer of target_species
        # TODO: check for mass based (not mole) get_material_flow_terms, but ok under mcas_prop_pack
        @self.Constraint(
            self.flowsheet().config.time,
            self.target_species,
            doc="mass transfer for adsorbed solutes in 'target_species' within 'gac_removed' (out of the bed)",
        )
        def eq_mass_transfer_cv(b, t, j):
            return (1 - b.conc_ratio_avg) * b.process_flow.properties_in[
                t
            ].get_material_flow_terms("Liq", j) == (
                -b.process_flow.mass_transfer_term[t, "Liq", j]
            )

        # mass balance for port not in control volume
        @self.Constraint(
            self.flowsheet().config.time,
            self.target_species,
            doc="mass balance for port not in control volume",
        )
        def eq_mass_transfer_port(b, t, j):
            return (
                b.gac_removed[t].get_material_flow_terms("Liq", j)
                == -b.process_flow.mass_transfer_term[t, "Liq", j]
            )

        # no mass transfer of inert_species, fix to 0 to avoid 0 solver tolerances
        for j in self.inert_species:
            self.process_flow.mass_transfer_term[:, "Liq", j].fix(0)
            self.gac_removed[0].get_material_flow_terms("Liq", j).fix(0)

        # ---------------------------------------------------------------------
        if (
            self.config.film_transfer_coefficient_type
            == FilmTransferCoefficientType.calculated
        ):

            self.N_Re = Var(
                initialize=10,
                bounds=(0, None),
                domain=NonNegativeReals,
                units=pyunits.dimensionless,
                doc="Reynolds number, correlations using Reynolds number valid in Re < 2e4",
            )
            self.N_Sc = Var(
                initialize=2000,
                bounds=(0, None),
                domain=NonNegativeReals,
                units=pyunits.dimensionless,
                doc="Schmidt number, correlations using Schmidt number valid in 0.7 < Sc < 1e4",
            )
            self.shape_correction_factor = Var(
                initialize=1,
                bounds=(0, None),
                domain=NonNegativeReals,
                units=pyunits.dimensionless,
                doc="shape correction factor",
            )

            @self.Constraint(
                self.flowsheet().config.time,
                doc="Reynolds number calculation",
            )
            def eq_reynolds_number(b, t):
                return (
                    b.N_Re * b.process_flow.properties_in[t].visc_d_phase["Liq"]
                    == b.process_flow.properties_in[t].dens_mass_phase["Liq"]
                    * b.particle_dia
                    * b.velocity_int
                )

            @self.Constraint(
                self.flowsheet().config.time,
                self.target_species,
                doc="Schmidt number calculation",
            )
            def eq_schmidt_number(b, t, j):
                return (
                    b.N_Sc
                    * b.process_flow.properties_in[t].dens_mass_phase["Liq"]
                    * b.process_flow.properties_in[t].diffus_phase_comp["Liq", j]
                    == b.process_flow.properties_in[t].visc_d_phase["Liq"]
                )

            @self.Constraint(
                self.flowsheet().config.time,
                self.target_species,
                doc="liquid phase film transfer rate from the Gnielinski correlation",
            )
            def eq_film_transfer_rate(b, t, j):
                return 1 == (b.kf * b.particle_dia) / (
                    b.shape_correction_factor
                    * (1 + 1.5 * (1 - b.bed_voidage))
                    * b.process_flow.properties_in[t].diffus_phase_comp["Liq", j]
                    * (2 + 0.644 * (b.N_Re**0.5) * (b.N_Sc ** (1 / 3)))
                )

        # ---------------------------------------------------------------------
        if (
            self.config.surface_diffusion_coefficient_type
            == SurfaceDiffusionCoefficientType.calculated
        ):

            self.particle_porosity = Var(
                initialize=0.65,
                bounds=(0, None),
                domain=NonNegativeReals,
                units=pyunits.dimensionless,
                doc="gac particle porosity",
            )
            self.tort = Var(
                initialize=1,
                bounds=(0, None),
                domain=NonNegativeReals,
                units=pyunits.dimensionless,
                doc="tortuosity of the path that the adsorbate must take as compared to the radius",
            )
            self.spdfr = Var(
                initialize=1,
                bounds=(0, None),
                domain=NonNegativeReals,
                units=pyunits.dimensionless,
                doc="surface-to-pore diffusion flux ratio",
            )

            @self.Constraint(
                self.flowsheet().config.time,
                self.target_species,
                doc="surface diffusion parameter",
            )
            def eq_surface_diffusion_coefficient_calculated(b, t, j):
                return (
                    b.ds * b.tort * b.equil_conc * b.particle_dens_app
                    == b.spdfr
                    * b.process_flow.properties_in[t].diffus_phase_comp["Liq", j]
                    * b.particle_porosity
                    * b.process_flow.properties_in[t].conc_mass_phase_comp["Liq", j]
                )

    # ---------------------------------------------------------------------
    # initialize method
    def initialize_build(
        self,
        state_args=None,
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
    ):
        """
        General wrapper for initialization routines

        Keyword Arguments:
            state_args : a dict of arguments to be passed to the property
                         package(s) to provide an initial state for
                         initialization (see documentation of the specific
                         property package) (default = {}).
            outlvl : sets output level of initialization routine
            optarg : solver options dictionary object (default=None)
            solver : str indicating which solver to use during
                     initialization (default = None)

        Returns: None
        """
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="unit")
        # set solver options
        opt = get_solver(solver, optarg)

        # ---------------------------------------------------------------------
        # set state_args from inlet state

        if state_args is None:
            state_args = {}
            state_dict = self.process_flow.properties_in[
                self.flowsheet().config.time.first()
            ].define_port_members()

            for k in state_dict.keys():
                if state_dict[k].is_indexed():
                    state_args[k] = {}
                    for m in state_dict[k].keys():
                        state_args[k][m] = state_dict[k][m].value
                else:
                    state_args[k] = state_dict[k].value

        # initialize control volume
        flags = self.process_flow.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=state_args,
        )

        init_log.info_high("Initialization Step 1 Complete.")
        # ---------------------------------------------------------------------
        # initialize gac_removed port

        # all inert species initialized to 0
        for j in self.inert_species:
            state_args["flow_mol_phase_comp"][("Liq", j)] = 0

        self.gac_removed.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=state_args,
        )

        init_log.info_high("Initialization Step 2 Complete.")
        # --------------------------------------------------------------------
        # solve unit

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(self, tee=slc.tee)

        init_log.info_high("Initialization Step 3 {}.".format(idaeslog.condition(res)))
        # ---------------------------------------------------------------------
        # release inlet state

        self.process_flow.release_state(flags, outlvl + 1)
        init_log.info("Initialization Complete: {}".format(idaeslog.condition(res)))

        if not check_optimal_termination(res):
            raise InitializationError(f"Unit model {self.name} failed to initialize")

    # ---------------------------------------------------------------------

    def _get_performance_contents(self, time_point=0):

        var_dict = {}

        # unit model variables
        var_dict["Freundlich isotherm k parameter"] = self.freund_k
        var_dict["Freundlich isotherm 1/n parameter"] = self.freund_ninv
        var_dict["Surface diffusion coefficient"] = self.ds
        var_dict["liquid phase film transfer coefficient"] = self.kf
        var_dict["superficial velocity"] = self.velocity_sup
        var_dict["bed void fraction"] = self.bed_voidage
        var_dict["bed length"] = self.bed_length
        var_dict["bed diameter"] = self.bed_diameter
        var_dict["bed volume"] = self.bed_volume
        var_dict["empty bed contact time"] = self.ebct
        var_dict["fluid residence time in the bed"] = self.residence_time
        var_dict["mass of fresh gac in the bed"] = self.bed_mass_gac
        var_dict["concentration ratio at operational time"] = self.conc_ratio_replace
        var_dict["time for the start of breakthrough"] = self.ele_operational_time[1]
        var_dict["operational time of the bed from fresh"] = self.operational_time
        var_dict["bed volumes treated"] = self.bed_volumes_treated
        var_dict["steady state average concentration ratio"] = self.conc_ratio_avg
        var_dict["gac usage/replacement/regeneration rate"] = self.gac_usage_rate

        # loop through desired state block properties indexed by [phase, comp]
        phase_comp_prop_dict = {
            "flow_mol_phase_comp": "molar flow rate",
            "conc_mass_phase_comp": "mass concentration",
        }
        for prop_name, prop_str in phase_comp_prop_dict.items():
            for j in self.config.property_package.component_list:
                if self.process_flow.properties_in[time_point].is_property_constructed(
                    prop_name
                ):
                    var_dict[f"{prop_str} of {j} @ process inlet"] = getattr(
                        self.process_flow.properties_in[time_point], prop_name
                    )["Liq", j]
                if self.process_flow.properties_out[time_point].is_property_constructed(
                    prop_name
                ):
                    var_dict[f"{prop_str} of {j} @ process outlet"] = getattr(
                        self.process_flow.properties_out[time_point], prop_name
                    )["Liq", j]
                if self.gac_removed[time_point].is_property_constructed(prop_name):
                    var_dict[f"{prop_str} of {j} @ gac removed outlet"] = getattr(
                        self.gac_removed[time_point], prop_name
                    )["Liq", j]

        # loop through desired state block properties indexed by [phase]
        phase_prop_dict = {
            "flow_vol_phase": "volumetric flow rate",
        }
        for prop_name, prop_str in phase_prop_dict.items():
            if self.process_flow.properties_in[time_point].is_property_constructed(
                prop_name
            ):
                var_dict[f"{prop_str} @ process inlet"] = getattr(
                    self.process_flow.properties_in[time_point], prop_name
                )["Liq"]
            if self.process_flow.properties_out[time_point].is_property_constructed(
                prop_name
            ):
                var_dict[f"{prop_str} @ process outlet"] = getattr(
                    self.process_flow.properties_out[time_point], prop_name
                )["Liq"]
            if self.gac_removed[time_point].is_property_constructed(prop_name):
                var_dict[f"{prop_str} @ gac removed outlet"] = getattr(
                    self.gac_removed[time_point], prop_name
                )["Liq"]

        return {"vars": var_dict}

    # ---------------------------------------------------------------------
    def _get_stream_table_contents(self, time_point=0):
        return create_stream_table_dataframe(
            {
                "Process Inlet": self.inlet,
                "Process Outlet": self.outlet,
                "GAC Removed": self.adsorbed,
            },
            time_point=time_point,
        )

    # ---------------------------------------------------------------------
    def calculate_scaling_factors(self):

        super().calculate_scaling_factors()

        # ---------------------------------------------------------------------
        # get scaling factors for target species and water to use for other variables

        for j in self.target_species:
            sf_solute = iscale.get_scaling_factor(
                self.process_flow.properties_in[0].flow_mol_phase_comp["Liq", j],
                default=1e4,  # default based on typical concentration for treatment
                warning=True,
            )

        for j in self.config.property_package.solvent_set:
            sf_solvent = iscale.get_scaling_factor(
                self.process_flow.properties_in[0].flow_mol_phase_comp["Liq", j],
                default=1e-3,  # default based on typical concentration for treatment
                warning=True,
            )

        # sf_volume based on magnitude of 0.018 water mw and 1000 dens
        sf_volume = sf_solvent * (100 / 0.01)
        sf_conc = sf_solute / sf_volume

        # ---------------------------------------------------------------------
        # overwrite default scaling for state block variables

        for j in self.target_species:
            iscale.set_scaling_factor(
                self.process_flow.properties_out[0].flow_mol_phase_comp["Liq", j],
                10 * sf_solute,
            )

        for j in self.inert_species:
            iscale.set_scaling_factor(
                self.gac_removed[0].flow_mol_phase_comp["Liq", j],
                1,
            )  # ensure lower concentration of zero flow components, below zero tol

        # dens stays at 1000 even though water flow is zero, using sf based on 0.1 assumed mw and 1000 dens
        if self.gac_removed[0].is_property_constructed("flow_vol_phase"):
            iscale.set_scaling_factor(
                self.gac_removed[0].flow_vol_phase["Liq"],
                sf_solute * (1000 / 0.1),
            )

        # ---------------------------------------------------------------------
        # scaling for gac created variables that are flow magnitude dependent

        if iscale.get_scaling_factor(self.equil_conc) is None:
            iscale.set_scaling_factor(self.equil_conc, sf_conc * 1e-2)

        if iscale.get_scaling_factor(self.bed_diameter) is None:
            iscale.set_scaling_factor(self.bed_diameter, sf_volume * 1e-2)

        if iscale.get_scaling_factor(self.bed_area) is None:
            iscale.set_scaling_factor(self.bed_area, sf_volume * 1e-2)

        if iscale.get_scaling_factor(self.bed_volume) is None:
            iscale.set_scaling_factor(self.bed_volume, sf_volume * 1e-2)

        if iscale.get_scaling_factor(self.bed_mass_gac) is None:
            iscale.set_scaling_factor(self.bed_mass_gac, sf_volume * 1e-6)

        if iscale.get_scaling_factor(self.mass_adsorbed) is None:
            iscale.set_scaling_factor(self.mass_adsorbed, sf_solute * 1e-6)

        if iscale.get_scaling_factor(self.gac_usage_rate) is None:
            iscale.set_scaling_factor(self.gac_usage_rate, sf_volume)

        # ---------------------------------------------------------------------
        # scaling for gac created variables that are flow magnitude independent

        if iscale.get_scaling_factor(self.freund_k) is None:
            iscale.set_scaling_factor(self.freund_k, 1)

        if iscale.get_scaling_factor(self.freund_ninv) is None:
            iscale.set_scaling_factor(self.freund_ninv, 1)

        if iscale.get_scaling_factor(self.ds) is None:
            iscale.set_scaling_factor(self.ds, 1e14)

        if iscale.get_scaling_factor(self.kf) is None:
            iscale.set_scaling_factor(self.kf, 1e5)

        if iscale.get_scaling_factor(self.dg) is None:
            iscale.set_scaling_factor(self.dg, 1e-4)

        if iscale.get_scaling_factor(self.N_Bi) is None:
            iscale.set_scaling_factor(self.N_Bi, 1)

        if iscale.get_scaling_factor(self.velocity_sup) is None:
            iscale.set_scaling_factor(self.velocity_sup, 1e3)

        if iscale.get_scaling_factor(self.velocity_int) is None:
            iscale.set_scaling_factor(self.velocity_int, 1e3)

        if iscale.get_scaling_factor(self.bed_voidage) is None:
            iscale.set_scaling_factor(self.bed_voidage, 1e1)

        if iscale.get_scaling_factor(self.bed_length) is None:
            iscale.set_scaling_factor(self.bed_length, 1)

        if iscale.get_scaling_factor(self.ebct) is None:
            iscale.set_scaling_factor(self.ebct, 1e-2)

        if iscale.get_scaling_factor(self.residence_time) is None:
            iscale.set_scaling_factor(self.residence_time, 1e-2)

        if iscale.get_scaling_factor(self.particle_dens_app) is None:
            iscale.set_scaling_factor(self.particle_dens_app, 1e-2)

        if iscale.get_scaling_factor(self.particle_dens_bulk) is None:
            iscale.set_scaling_factor(self.particle_dens_bulk, 1e-2)

        if iscale.get_scaling_factor(self.particle_dia) is None:
            iscale.set_scaling_factor(self.particle_dia, 1e3)

        iscale.set_scaling_factor(self.a0, 1)
        iscale.set_scaling_factor(self.a1, 1)
        iscale.set_scaling_factor(self.b0, 1)
        iscale.set_scaling_factor(self.b1, 1)
        iscale.set_scaling_factor(self.b2, 1)
        iscale.set_scaling_factor(self.b3, 10)
        iscale.set_scaling_factor(self.b4, 1)

        if iscale.get_scaling_factor(self.min_N_St) is None:
            iscale.set_scaling_factor(self.min_N_St, 1e-1)

        if iscale.get_scaling_factor(self.min_ebct) is None:
            iscale.set_scaling_factor(self.min_ebct, 1e-2)

        if iscale.get_scaling_factor(self.throughput) is None:
            iscale.set_scaling_factor(self.throughput, 1)

        if iscale.get_scaling_factor(self.min_residence_time) is None:
            iscale.set_scaling_factor(self.min_residence_time, 1e-2)

        if iscale.get_scaling_factor(self.min_operational_time) is None:
            iscale.set_scaling_factor(self.min_operational_time, 1e-6)

        if iscale.get_scaling_factor(self.conc_ratio_replace) is None:
            iscale.set_scaling_factor(self.conc_ratio_replace, 1e1)

        if iscale.get_scaling_factor(self.operational_time) is None:
            iscale.set_scaling_factor(self.operational_time, 1e-6)

        if iscale.get_scaling_factor(self.bed_volumes_treated) is None:
            iscale.set_scaling_factor(self.bed_volumes_treated, 1e-5)

        for ele in range(1, self.elements_ss_approx.value + 1):

            if iscale.get_scaling_factor(self.ele_throughput[ele]) is None:
                iscale.set_scaling_factor(self.ele_throughput[ele], 1)

            if iscale.get_scaling_factor(self.ele_min_operational_time[ele]) is None:
                iscale.set_scaling_factor(self.ele_min_operational_time[ele], 1e-6)

            if iscale.get_scaling_factor(self.ele_conc_ratio_term[ele]) is None:
                iscale.set_scaling_factor(self.ele_conc_ratio_term[ele], 1e2)

        for ele in range(self.elements_ss_approx.value + 1):

            if iscale.get_scaling_factor(self.ele_conc_ratio_replace[ele]) is None:
                iscale.set_scaling_factor(self.ele_conc_ratio_replace[ele], 1e1)

            if iscale.get_scaling_factor(self.ele_operational_time[ele]) is None:
                iscale.set_scaling_factor(self.ele_operational_time[ele], 1e-6)

        if iscale.get_scaling_factor(self.conc_ratio_avg) is None:
            iscale.set_scaling_factor(self.conc_ratio_avg, 1e1)

        if hasattr(self, "N_Re"):
            if iscale.get_scaling_factor(self.N_Re) is None:
                iscale.set_scaling_factor(self.N_Re, 1)

        if hasattr(self, "N_Sc"):
            if iscale.get_scaling_factor(self.N_Sc) is None:
                iscale.set_scaling_factor(self.N_Sc, 1e-3)

        if hasattr(self, "shape_correction_factor"):
            if iscale.get_scaling_factor(self.shape_correction_factor) is None:
                iscale.set_scaling_factor(self.shape_correction_factor, 1)

        if hasattr(self, "particle_porosity"):
            if iscale.get_scaling_factor(self.particle_porosity) is None:
                iscale.set_scaling_factor(self.particle_porosity, 1e1)

        if hasattr(self, "tort"):
            if iscale.get_scaling_factor(self.tort) is None:
                iscale.set_scaling_factor(self.tort, 1)

        if hasattr(self, "spdfr"):
            if iscale.get_scaling_factor(self.spdfr) is None:
                iscale.set_scaling_factor(self.spdfr, 1)

        # (optional) transforming constraints

    @property
    def default_costing_method(self):
        return cost_gac



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

"""
Stoichiometric Softening pretreatment process

This will build a stoichiometric pretreatment process unit by unit. The
overall process is diagrammed below::

                    Lime stream
                        |
                        V
    inlet stream ---> [Mixer] --- outlet stream ---> [StoichiometricReactor] ---> mixed flow ... (see below)


    ...mixed flow ---> [Separator] ---> exit stream (to RO)
                           |
                           V
                       waste stream

Stoich Reactions::

     Ca(HCO3)2 +   Ca(OH)2 --> 2 CaCO3 +  2 H2O
     Mg(HCO3)2 + 2 Ca(OH)2 --> 2 CaCO3 + Mg(OH)2 +  2 H2O
"""

# Importing the object for units from pyomo
from pyomo.environ import units as pyunits, Expression, NonNegativeReals

# Imports from idaes core
from idaes.core import AqueousPhase
from idaes.core.base.components import Solvent, Solute, Anion
from idaes.core.base.phases import PhaseType as PT
from idaes.core.util.math import smooth_min

# Imports from idaes generic models
import idaes.models.properties.modular_properties.pure.Perrys as Perrys
from idaes.models.properties.modular_properties.pure.ConstantProperties import Constant
from idaes.models.properties.modular_properties.state_definitions import FTPx
from idaes.models.properties.modular_properties.eos.ideal import Ideal
from idaes.models.properties.modular_properties.reactions.rate_constant import arrhenius
from idaes.models.properties.modular_properties.reactions.rate_forms import (
    power_law_rate,
)

# Importing the enum for concentration unit basis used in the 'get_concentration_term' function
from idaes.models.properties.modular_properties.base.generic_reaction import (
    ConcentrationForm,
)

# Import the object/function for heat of reaction
from idaes.models.properties.modular_properties.reactions.dh_rxn import constant_dh_rxn

# Import specific pyomo objects
from pyomo.environ import (
    ConcreteModel,
    Var,
    Constraint,
    TransformationFactory,
    value,
)

from pyomo.network import Arc
from idaes.core.util.initialization import propagate_state

from watertap.examples.flowsheets.full_treatment_train.util import (
    solve_block,
    check_dof,
)

from idaes.core.solvers import get_solver

# Import the idaes objects for Generic Properties and Reactions
from idaes.models.properties.modular_properties.base.generic_property import (
    GenericParameterBlock,
)
from idaes.models.properties.modular_properties.base.generic_reaction import (
    GenericReactionParameterBlock,
)

# Import the idaes object for the StoichiometricReactor unit model
from idaes.models.unit_models.stoichiometric_reactor import (
    StoichiometricReactor,
)

# Import the Mixer and Separator unit model
from idaes.models.unit_models import (
    Separator,
    Mixer,
)
from watertap.examples.flowsheets.full_treatment_train.flowsheet_components import (
    costing,
)

from idaes.models.unit_models.separator import (
    SplittingType,
    EnergySplittingType,
)

# Import the core idaes objects for Flowsheets and types of balances
from idaes.core import FlowsheetBlock

# Import log10 function from pyomo
import idaes.logger as idaeslog

# Grab the scaling utilities

from watertap.examples.flowsheets.full_treatment_train.chemical_flowsheet_util import (
    set_H2O_molefraction,
    zero_out_non_H2O_molefractions,
    fix_all_molefractions,
    unfix_all_molefractions,
)

__author__ = "Austin Ladshaw, Srikanth Allu"

# Configuration dictionary
stoich_softening_thermo_config = {
    "components": {
        "H2O": {
            "type": Solvent,
            # Define the methods used to calculate the following properties
            "dens_mol_liq_comp": Perrys,
            "enth_mol_liq_comp": Perrys,
            "cp_mol_liq_comp": Perrys,
            "entr_mol_liq_comp": Perrys,
            # Parameter data is always associated with the methods defined above
            "parameter_data": {
                "mw": (18.0153, pyunits.g / pyunits.mol),
                "pressure_crit": (220.64e5, pyunits.Pa),
                "temperature_crit": (647, pyunits.K),
                # Comes from Perry's Handbook:  p. 2-98
                "dens_mol_liq_comp_coeff": {
                    "eqn_type": 1,
                    "1": (5.459, pyunits.kmol * pyunits.m**-3),
                    "2": (0.30542, pyunits.dimensionless),
                    "3": (647.13, pyunits.K),
                    "4": (0.081, pyunits.dimensionless),
                },
                "enth_mol_form_liq_comp_ref": (-285.830, pyunits.kJ / pyunits.mol),
                "enth_mol_form_vap_comp_ref": (0, pyunits.kJ / pyunits.mol),
                # Comes from Perry's Handbook:  p. 2-174
                "cp_mol_liq_comp_coeff": {
                    "1": (2.7637e5, pyunits.J / pyunits.kmol / pyunits.K),
                    "2": (-2.0901e3, pyunits.J / pyunits.kmol / pyunits.K**2),
                    "3": (8.125, pyunits.J / pyunits.kmol / pyunits.K**3),
                    "4": (-1.4116e-2, pyunits.J / pyunits.kmol / pyunits.K**4),
                    "5": (9.3701e-6, pyunits.J / pyunits.kmol / pyunits.K**5),
                },
                "cp_mol_ig_comp_coeff": {
                    "A": (30.09200, pyunits.J / pyunits.mol / pyunits.K),
                    "B": (
                        6.832514,
                        pyunits.J
                        * pyunits.mol**-1
                        * pyunits.K**-1
                        * pyunits.kiloK**-1,
                    ),
                    "C": (
                        6.793435,
                        pyunits.J
                        * pyunits.mol**-1
                        * pyunits.K**-1
                        * pyunits.kiloK**-2,
                    ),
                    "D": (
                        -2.534480,
                        pyunits.J
                        * pyunits.mol**-1
                        * pyunits.K**-1
                        * pyunits.kiloK**-3,
                    ),
                    "E": (
                        0.082139,
                        pyunits.J
                        * pyunits.mol**-1
                        * pyunits.K**-1
                        * pyunits.kiloK**2,
                    ),
                    "F": (-250.8810, pyunits.kJ / pyunits.mol),
                    "G": (223.3967, pyunits.J / pyunits.mol / pyunits.K),
                    "H": (0, pyunits.kJ / pyunits.mol),
                },
                "entr_mol_form_liq_comp_ref": (
                    69.95,
                    pyunits.J / pyunits.K / pyunits.mol,
                ),
                "pressure_sat_comp_coeff": {
                    "A": (4.6543, None),  # [1], temperature range 255.9 K - 373 K
                    "B": (1435.264, pyunits.K),
                    "C": (-64.848, pyunits.K),
                },
            },
            # End parameter_data
        },
        "Ca(OH)2": {
            "type": Solute,
            "valid_phase_types": PT.aqueousPhase,
            # Define the methods used to calculate the following properties
            "dens_mol_liq_comp": Constant,
            "enth_mol_liq_comp": Constant,
            "cp_mol_liq_comp": Constant,
            "entr_mol_liq_comp": Constant,
            # Parameter data is always associated with the methods defined above
            "parameter_data": {
                "mw": (74.093, pyunits.g / pyunits.mol),
                "dens_mol_liq_comp_coeff": (55, pyunits.kmol * pyunits.m**-3),
                "enth_mol_form_liq_comp_ref": (-986, pyunits.kJ / pyunits.mol),
                "cp_mol_liq_comp_coeff": (75348, pyunits.J / pyunits.kmol / pyunits.K),
                "entr_mol_form_liq_comp_ref": (83, pyunits.J / pyunits.K / pyunits.mol),
            },
            # End parameter_data
        },
        "NaCl": {
            "type": Solute,
            "valid_phase_types": PT.aqueousPhase,
            # Define the methods used to calculate the following properties
            "dens_mol_liq_comp": Constant,
            "enth_mol_liq_comp": Constant,
            "cp_mol_liq_comp": Constant,
            "entr_mol_liq_comp": Constant,
            # Parameter data is always associated with the methods defined above
            "parameter_data": {
                "mw": (58.44, pyunits.g / pyunits.mol),
                "dens_mol_liq_comp_coeff": (55, pyunits.kmol * pyunits.m**-3),
                "enth_mol_form_liq_comp_ref": (-407, pyunits.kJ / pyunits.mol),
                "cp_mol_liq_comp_coeff": (75348, pyunits.J / pyunits.kmol / pyunits.K),
                "entr_mol_form_liq_comp_ref": (
                    115,
                    pyunits.J / pyunits.K / pyunits.mol,
                ),
            },
            # End parameter_data
        },
        "Cl_-": {
            "type": Anion,
            "charge": -1,
            # Define the methods used to calculate the following properties
            "dens_mol_liq_comp": Constant,
            "enth_mol_liq_comp": Constant,
            "cp_mol_liq_comp": Constant,
            "entr_mol_liq_comp": Constant,
            # Parameter data is always associated with the methods defined above
            "parameter_data": {
                "mw": (35.453, pyunits.g / pyunits.mol),
                "dens_mol_liq_comp_coeff": (55, pyunits.kmol * pyunits.m**-3),
                "enth_mol_form_liq_comp_ref": (-407, pyunits.kJ / pyunits.mol),
                "cp_mol_liq_comp_coeff": (75348, pyunits.J / pyunits.kmol / pyunits.K),
                "entr_mol_form_liq_comp_ref": (
                    115,
                    pyunits.J / pyunits.K / pyunits.mol,
                ),
            },
            # End parameter_data
        },
        "CaCO3": {
            "type": Solute,
            "valid_phase_types": PT.aqueousPhase,
            # Define the methods used to calculate the following properties
            "dens_mol_liq_comp": Constant,
            "enth_mol_liq_comp": Constant,
            "cp_mol_liq_comp": Constant,
            "entr_mol_liq_comp": Constant,
            # Parameter data is always associated with the methods defined above
            "parameter_data": {
                "mw": (100.09, pyunits.g / pyunits.mol),
                "dens_mol_liq_comp_coeff": (55, pyunits.kmol * pyunits.m**-3),
                "enth_mol_form_liq_comp_ref": (-1207, pyunits.kJ / pyunits.mol),
                "cp_mol_liq_comp_coeff": (75348, pyunits.J / pyunits.kmol / pyunits.K),
                "entr_mol_form_liq_comp_ref": (92, pyunits.J / pyunits.K / pyunits.mol),
            },
            # End parameter_data
        },
        "Ca(HCO3)2": {
            "type": Solute,
            "valid_phase_types": PT.aqueousPhase,
            # Define the methods used to calculate the following properties
            "dens_mol_liq_comp": Constant,
            "enth_mol_liq_comp": Constant,
            "cp_mol_liq_comp": Constant,
            "entr_mol_liq_comp": Constant,
            # Parameter data is always associated with the methods defined above
            "parameter_data": {
                "mw": (162.11, pyunits.g / pyunits.mol),
                "dens_mol_liq_comp_coeff": (55, pyunits.kmol * pyunits.m**-3),
                "enth_mol_form_liq_comp_ref": (-542, pyunits.kJ / pyunits.mol),
                "cp_mol_liq_comp_coeff": (75348, pyunits.J / pyunits.kmol / pyunits.K),
                "entr_mol_form_liq_comp_ref": (
                    -53,
                    pyunits.J / pyunits.K / pyunits.mol,
                ),
            },
            # End parameter_data
        },
        "Mg(OH)2": {
            "type": Solute,
            "valid_phase_types": PT.aqueousPhase,
            # Define the methods used to calculate the following properties
            "dens_mol_liq_comp": Constant,
            "enth_mol_liq_comp": Constant,
            "cp_mol_liq_comp": Constant,
            "entr_mol_liq_comp": Constant,
            # Parameter data is always associated with the methods defined above
            "parameter_data": {
                "mw": (74.093, pyunits.g / pyunits.mol),
                "dens_mol_liq_comp_coeff": (55, pyunits.kmol * pyunits.m**-3),
                "enth_mol_form_liq_comp_ref": (-924, pyunits.kJ / pyunits.mol),
                "cp_mol_liq_comp_coeff": (75348, pyunits.J / pyunits.kmol / pyunits.K),
                "entr_mol_form_liq_comp_ref": (63, pyunits.J / pyunits.K / pyunits.mol),
            },
            # End parameter_data
        },
        "Mg(HCO3)2": {
            "type": Solute,
            "valid_phase_types": PT.aqueousPhase,
            # Define the methods used to calculate the following properties
            "dens_mol_liq_comp": Constant,
            "enth_mol_liq_comp": Constant,
            "cp_mol_liq_comp": Constant,
            "entr_mol_liq_comp": Constant,
            # Parameter data is always associated with the methods defined above
            "parameter_data": {
                "mw": (146.34, pyunits.g / pyunits.mol),
                "dens_mol_liq_comp_coeff": (55, pyunits.kmol * pyunits.m**-3),
                "enth_mol_form_liq_comp_ref": (-221, pyunits.kJ / pyunits.mol),
                "cp_mol_liq_comp_coeff": (75348, pyunits.J / pyunits.kmol / pyunits.K),
                "entr_mol_form_liq_comp_ref": (
                    -74,
                    pyunits.J / pyunits.K / pyunits.mol,
                ),
            },
            # End parameter_data
        },
        "SO4_2-": {
            "type": Anion,
            "charge": -2,
            # Define the methods used to calculate the following properties
            "dens_mol_liq_comp": Constant,
            "enth_mol_liq_comp": Constant,
            "cp_mol_liq_comp": Constant,
            "entr_mol_liq_comp": Constant,
            # Parameter data is always associated with the methods defined above
            "parameter_data": {
                "mw": (96.07, pyunits.g / pyunits.mol),
                "dens_mol_liq_comp_coeff": (55, pyunits.kmol * pyunits.m**-3),
                "enth_mol_form_liq_comp_ref": (-909, pyunits.kJ / pyunits.mol),
                "cp_mol_liq_comp_coeff": (75348, pyunits.J / pyunits.kmol / pyunits.K),
                "entr_mol_form_liq_comp_ref": (20, pyunits.J / pyunits.K / pyunits.mol),
            },
            # End parameter_data
        },
    },
    # End Component list
    "phases": {
        "Liq": {"type": AqueousPhase, "equation_of_state": Ideal},
    },
    "state_definition": FTPx,
    "state_bounds": {
        "flow_mol": (0, 50, 100),
        "temperature": (273.15, 300, 650),
        "pressure": (5e4, 1e5, 1e6),
    },
    "pressure_ref": 1e5,
    "temperature_ref": 300,
    "base_units": {
        "time": pyunits.s,
        "length": pyunits.m,
        "mass": pyunits.kg,
        "amount": pyunits.mol,
        "temperature": pyunits.K,
    },
    "default_scaling_factors": {
        ("mole_frac_comp", "Ca(HCO3)2"): 1e4,
        ("mole_frac_comp", "Ca(OH)2"): 1e4,
        ("mole_frac_comp", "CaCO3"): 1e4,
        ("mole_frac_comp", "H2O"): 1,
        ("mole_frac_comp", "Mg(HCO3)2"): 1e3,
        ("mole_frac_comp", "Mg(OH)2"): 1e3,
        ("mole_frac_comp", "NaCl"): 1e2,
        ("mole_frac_comp", "SO4_2-"): 1e3,
        ("mole_frac_comp", "Cl_-"): 1e3,
        ("flow_mol", None): 1e-1,
        ("temperature", None): 1e-2,
        ("pressure", None): 1e-6,
        ("flow_mol_phase_comp", ("Liq", "Ca(HCO3)2")): 1e4 * 1e-1,
        ("flow_mol_phase_comp", ("Liq", "Ca(OH)2")): 1e4 * 1e-1,
        ("flow_mol_phase_comp", ("Liq", "CaCO3")): 1e4 * 1e-1,
        ("flow_mol_phase_comp", ("Liq", "H2O")): 1 * 1e-1,
        ("flow_mol_phase_comp", ("Liq", "Mg(HCO3)2")): 1e3 * 1e-1,
        ("flow_mol_phase_comp", ("Liq", "Mg(OH)2")): 1e3 * 1e-1,
        ("flow_mol_phase_comp", ("Liq", "NaCl")): 1e3 * 1e-1,
        ("flow_mol_phase_comp", ("Liq", "SO4_2-")): 1e3 * 1e-1,
        ("flow_mol_phase_comp", ("Liq", "Cl_-")): 1e3 * 1e-1,
        ("flow_mol_phase", "Liq"): 1e-1,
    },
}
# End softening_thermo_config definition

# This config is for stoich softening
stoich_softening_reaction_config = {
    "base_units": {
        "time": pyunits.s,
        "length": pyunits.m,
        "mass": pyunits.kg,
        "amount": pyunits.mol,
        "temperature": pyunits.K,
    },
    "rate_reactions": {
        "Ca_removal": {
            "stoichiometry": {
                ("Liq", "Ca(HCO3)2"): -1,
                ("Liq", "Ca(OH)2"): -1,
                ("Liq", "CaCO3"): 2,
                ("Liq", "H2O"): 2,
            },
            "heat_of_reaction": constant_dh_rxn,
            "rate_constant": arrhenius,
            "rate_form": power_law_rate,
            "concentration_form": ConcentrationForm.moleFraction,
            "parameter_data": {
                "arrhenius_const": (1, pyunits.mol / pyunits.m**3 / pyunits.s),
                "energy_activation": (0, pyunits.J / pyunits.mol),
                "dh_rxn_ref": (0, pyunits.J / pyunits.mol),
            },
        },
        "Mg_removal": {
            "stoichiometry": {
                ("Liq", "Mg(HCO3)2"): -1,
                ("Liq", "Ca(OH)2"): -2,
                ("Liq", "CaCO3"): 2,
                ("Liq", "Mg(OH)2"): 1,
                ("Liq", "H2O"): 2,
            },
            "heat_of_reaction": constant_dh_rxn,
            "rate_constant": arrhenius,
            "rate_form": power_law_rate,
            "concentration_form": ConcentrationForm.moleFraction,
            "parameter_data": {
                "arrhenius_const": (1, pyunits.mol / pyunits.m**3 / pyunits.s),
                "energy_activation": (0, pyunits.J / pyunits.mol),
                "dh_rxn_ref": (0, pyunits.J / pyunits.mol),
            },
        },
    },
}
# End reaction_config definition

# Get default solver for testing
solver = get_solver()


def build_stoich_softening_prop(model):
    model.fs.stoich_softening_thermo_params = GenericParameterBlock(
        **stoich_softening_thermo_config
    )
    model.fs.stoich_softening_rxn_params = GenericReactionParameterBlock(
        property_package=model.fs.stoich_softening_thermo_params,
        **stoich_softening_reaction_config
    )


def build_stoich_softening_mixer_unit(model):
    model.fs.stoich_softening_mixer_unit = Mixer(
        property_package=model.fs.stoich_softening_thermo_params,
        inlet_list=["inlet_stream", "lime_stream"],
    )

    # add new constraint for dosing rate
    dr = value(
        model.fs.stoich_softening_mixer_unit.lime_stream.flow_mol[0]
        * model.fs.stoich_softening_mixer_unit.lime_stream.mole_frac_comp[0, "Ca(OH)2"]
        * model.fs.stoich_softening_thermo_params.get_component("Ca(OH)2").mw
    )
    model.fs.stoich_softening_mixer_unit.dosing_rate = Var(
        initialize=dr,
        domain=NonNegativeReals,
        bounds=(0.0, 1),
        units=pyunits.kg / pyunits.s,
    )

    def _dosing_rate_cons(blk):
        return (
            blk.dosing_rate
            == blk.lime_stream.flow_mol[0]
            * blk.lime_stream.mole_frac_comp[0, "Ca(OH)2"]
            * blk.lime_stream_state[0].params.get_component("Ca(OH)2").mw
        )

    model.fs.stoich_softening_mixer_unit.dosing_cons = Constraint(
        rule=_dosing_rate_cons
    )


def build_stoich_softening_reactor_unit(
    model, frac_excess_lime=0.01, frac_used_for_Ca_removal=0.99
):
    model.fs.stoich_softening_reactor_unit = StoichiometricReactor(
        property_package=model.fs.stoich_softening_thermo_params,
        reaction_package=model.fs.stoich_softening_rxn_params,
        has_heat_transfer=False,
        has_heat_of_reaction=False,
        has_pressure_change=False,
    )
    set_stoich_softening_reactor_extents(
        model, frac_excess_lime, frac_used_for_Ca_removal
    )


def build_stoich_softening_separator_unit(model, solids_removal_frac=0.99):
    model.fs.stoich_softening_separator_unit = Separator(
        property_package=model.fs.stoich_softening_thermo_params,
        outlet_list=["waste_stream", "outlet_stream"],
        split_basis=SplittingType.componentFlow,
        energy_split_basis=EnergySplittingType.equal_temperature,
    )

    total_molar_density = (
        value(
            model.fs.stoich_softening_separator_unit.mixed_state[0.0].dens_mol_phase[
                "Liq"
            ]
        )
        / 1000
    )

    # add new constraint for hardness (with initial value guess)
    hrd = (
        model.fs.stoich_softening_separator_unit.mixed_state[0.0]
        .mole_frac_comp["Ca(HCO3)2"]
        .value
        + model.fs.stoich_softening_separator_unit.mixed_state[0.0]
        .mole_frac_comp["Mg(HCO3)2"]
        .value
    )
    hrd = hrd * 2 * 50000 * total_molar_density
    model.fs.stoich_softening_separator_unit.hardness = Var(
        initialize=hrd,
        domain=NonNegativeReals,
        # bounds=
        # units =
    )

    def _hardness_cons(blk):
        return blk.hardness == (
            blk.outlet_stream.mole_frac_comp[0, "Ca(HCO3)2"]
            + blk.outlet_stream.mole_frac_comp[0, "Mg(HCO3)2"]
            + blk.outlet_stream.mole_frac_comp[0, "Ca(OH)2"]
            + blk.outlet_stream.mole_frac_comp[0, "CaCO3"]
            + blk.outlet_stream.mole_frac_comp[0, "Mg(OH)2"]
        ) * 2 * 50 * (blk.mixed_state[0.0].dens_mol_phase["Liq"])

    model.fs.stoich_softening_separator_unit.hardness_cons = Constraint(
        rule=_hardness_cons
    )

    set_stoich_softening_separator_split_frac(
        model, solids_removal_frac=solids_removal_frac
    )
    fix_stoich_softening_separator_split_frac(model)


def set_stoich_softening_mixer_inlets(
    model,
    dosing_rate_of_lime_mg_per_s=25,
    inlet_water_density_kg_per_L=1,
    inlet_temperature_K=298,
    inlet_pressure_Pa=101325,
    inlet_flow_mol_per_s=10,
    inlet_total_hardness_mg_per_L=200,
    hardness_fraction_to_Ca=0.5,
    inlet_salinity_psu=35,
    inlet_excess_chlorine_ppm=35,
    inlet_sulfate_ppm=2000,
):

    # inlet stream
    model.fs.stoich_softening_mixer_unit.inlet_stream.flow_mol[0].set_value(
        inlet_flow_mol_per_s
    )
    model.fs.stoich_softening_mixer_unit.inlet_stream.pressure[0].set_value(
        inlet_pressure_Pa
    )
    model.fs.stoich_softening_mixer_unit.inlet_stream.temperature[0].set_value(
        inlet_temperature_K
    )

    zero_out_non_H2O_molefractions(model.fs.stoich_softening_mixer_unit.inlet_stream)
    # Calculate molefractions for Ca(HCO3)2 and Mg(HCO3)2
    total_molar_density = inlet_water_density_kg_per_L / 18 * 1000  # mol/L
    if hardness_fraction_to_Ca > 1:
        hardness_fraction_to_Ca = 1
    if hardness_fraction_to_Ca < 0:
        hardness_fraction_to_Ca = 0
    x_Ca = (
        inlet_total_hardness_mg_per_L / 50000 / 2 / total_molar_density
    ) * hardness_fraction_to_Ca
    x_Mg = (inlet_total_hardness_mg_per_L / 50000 / 2 / total_molar_density) * (
        1 - hardness_fraction_to_Ca
    )
    model.fs.stoich_softening_mixer_unit.inlet_stream.mole_frac_comp[
        0, "Ca(HCO3)2"
    ].set_value(x_Ca)
    model.fs.stoich_softening_mixer_unit.inlet_stream.mole_frac_comp[
        0, "Mg(HCO3)2"
    ].set_value(x_Mg)

    x_so4 = (
        inlet_sulfate_ppm * inlet_water_density_kg_per_L / 96060 / total_molar_density
    )
    model.fs.stoich_softening_mixer_unit.inlet_stream.mole_frac_comp[
        0, "SO4_2-"
    ].set_value(x_so4)

    x_cl = (
        inlet_excess_chlorine_ppm
        * inlet_water_density_kg_per_L
        / 96060
        / total_molar_density
    )
    model.fs.stoich_softening_mixer_unit.inlet_stream.mole_frac_comp[
        0, "Cl_-"
    ].set_value(x_cl)

    total_salt = (
        value(
            model.fs.stoich_softening_mixer_unit.inlet_stream.mole_frac_comp[
                0, "Ca(HCO3)2"
            ]
        )
        * total_molar_density
        * 101
    )
    total_salt += (
        value(
            model.fs.stoich_softening_mixer_unit.inlet_stream.mole_frac_comp[
                0, "Mg(HCO3)2"
            ]
        )
        * total_molar_density
        * 85.31
    )
    psu_from_hardness = total_salt / (total_molar_density * 18) * 1000
    if psu_from_hardness < inlet_salinity_psu:
        psu_from_nacl = inlet_salinity_psu - psu_from_hardness
        x_NaCl = (
            psu_from_nacl
            * (total_molar_density * 18)
            / 1000
            / total_molar_density
            / 58.44
        )
        model.fs.stoich_softening_mixer_unit.inlet_stream.mole_frac_comp[
            0, "NaCl"
        ].set_value(x_NaCl)

    set_H2O_molefraction(model.fs.stoich_softening_mixer_unit.inlet_stream)

    # lime stream
    model.fs.stoich_softening_mixer_unit.lime_stream.pressure[0].set_value(
        inlet_pressure_Pa
    )
    model.fs.stoich_softening_mixer_unit.lime_stream.temperature[0].set_value(
        inlet_temperature_K
    )
    # Use given dosing rate value to estimate OCl_- molefraction and flow rate for naocl stream
    zero_out_non_H2O_molefractions(model.fs.stoich_softening_mixer_unit.lime_stream)
    model.fs.stoich_softening_mixer_unit.lime_stream.mole_frac_comp[
        0, "Ca(OH)2"
    ].set_value(1)
    set_H2O_molefraction(model.fs.stoich_softening_mixer_unit.lime_stream)
    flow_of_lime = (
        dosing_rate_of_lime_mg_per_s
        / model.fs.stoich_softening_mixer_unit.lime_stream.mole_frac_comp[
            0, "Ca(OH)2"
        ].value
        / 74.44
        / 1000
    )
    model.fs.stoich_softening_mixer_unit.lime_stream.flow_mol[0].set_value(flow_of_lime)

    model.fs.stoich_softening_mixer_unit.dosing_rate.set_value(
        dosing_rate_of_lime_mg_per_s * 1e6
    )


def set_stoich_softening_reactor_inlets(
    model,
    dosage_of_lime_mg_per_L=140,
    inlet_water_density_kg_per_L=1,
    inlet_temperature_K=298,
    inlet_pressure_Pa=101325,
    inlet_flow_mol_per_s=10,
    inlet_total_hardness_mg_per_L=200,
    hardness_fraction_to_Ca=0.5,
    inlet_salinity_psu=35,
    inlet_excess_chlorine_ppm=35,
    inlet_sulfate_ppm=2000,
):

    # inlet stream
    model.fs.stoich_softening_reactor_unit.inlet.flow_mol[0].set_value(
        inlet_flow_mol_per_s
    )
    model.fs.stoich_softening_reactor_unit.inlet.pressure[0].set_value(
        inlet_pressure_Pa
    )
    model.fs.stoich_softening_reactor_unit.inlet.temperature[0].set_value(
        inlet_temperature_K
    )

    zero_out_non_H2O_molefractions(model.fs.stoich_softening_reactor_unit.inlet)
    # Calculate molefractions for Ca(HCO3)2 and Mg(HCO3)2
    total_molar_density = inlet_water_density_kg_per_L / 18 * 1000  # mol/L
    if hardness_fraction_to_Ca > 1:
        hardness_fraction_to_Ca = 1
    if hardness_fraction_to_Ca < 0:
        hardness_fraction_to_Ca = 0
    x_Ca = (
        inlet_total_hardness_mg_per_L / 50000 / 2 / total_molar_density
    ) * hardness_fraction_to_Ca
    x_Mg = (inlet_total_hardness_mg_per_L / 50000 / 2 / total_molar_density) * (
        1 - hardness_fraction_to_Ca
    )
    model.fs.stoich_softening_reactor_unit.inlet.mole_frac_comp[
        0, "Ca(HCO3)2"
    ].set_value(x_Ca)
    model.fs.stoich_softening_reactor_unit.inlet.mole_frac_comp[
        0, "Mg(HCO3)2"
    ].set_value(x_Mg)

    x_so4 = (
        inlet_sulfate_ppm * inlet_water_density_kg_per_L / 96060 / total_molar_density
    )
    model.fs.stoich_softening_reactor_unit.inlet.mole_frac_comp[0, "SO4_2-"].set_value(
        x_so4
    )

    x_cl = (
        inlet_excess_chlorine_ppm
        * inlet_water_density_kg_per_L
        / 96060
        / total_molar_density
    )
    model.fs.stoich_softening_reactor_unit.inlet.mole_frac_comp[0, "Cl_-"].set_value(
        x_cl
    )

    total_salt = (
        value(
            model.fs.stoich_softening_reactor_unit.inlet.mole_frac_comp[0, "Ca(HCO3)2"]
        )
        * total_molar_density
        * 101
    )
    total_salt += (
        value(
            model.fs.stoich_softening_reactor_unit.inlet.mole_frac_comp[0, "Mg(HCO3)2"]
        )
        * total_molar_density
        * 85.31
    )
    psu_from_hardness = total_salt / (total_molar_density * 18) * 1000
    if psu_from_hardness < inlet_salinity_psu:
        psu_from_nacl = inlet_salinity_psu - psu_from_hardness
        x_NaCl = (
            psu_from_nacl
            * (total_molar_density * 18)
            / 1000
            / total_molar_density
            / 58.44
        )
        model.fs.stoich_softening_reactor_unit.inlet.mole_frac_comp[
            0, "NaCl"
        ].set_value(x_NaCl)

    x_lime = dosage_of_lime_mg_per_L / 1000 / 74.093 / total_molar_density
    model.fs.stoich_softening_reactor_unit.inlet.mole_frac_comp[0, "Ca(OH)2"].set_value(
        x_lime
    )

    set_H2O_molefraction(model.fs.stoich_softening_reactor_unit.inlet)


def set_stoich_softening_reactor_extents(
    model, frac_excess_lime=0.01, frac_used_for_Ca_removal=0.99
):
    if frac_excess_lime > 1:
        frac_excess_lime = 1
    if frac_excess_lime < 0:
        frac_excess_lime = 0
    if frac_used_for_Ca_removal > 1:
        frac_used_for_Ca_removal = 1
    if frac_used_for_Ca_removal < 0:
        frac_used_for_Ca_removal = 0

    model.fs.stoich_softening_reactor_unit.frac_excess_lime = Var(
        initialize=frac_excess_lime
    )
    model.fs.stoich_softening_reactor_unit.frac_excess_lime.fix()

    model.fs.stoich_softening_reactor_unit.frac_used_for_Ca_removal = Var(
        initialize=frac_used_for_Ca_removal
    )
    model.fs.stoich_softening_reactor_unit.frac_used_for_Ca_removal.fix()

    def _ca_extent_cons(blk):
        a = (
            blk.inlet.mole_frac_comp[0, "Ca(OH)2"]
            * (1 - blk.frac_excess_lime)
            * blk.inlet.flow_mol[0]
            * blk.frac_used_for_Ca_removal
        )
        b = blk.inlet.mole_frac_comp[0, "Ca(HCO3)2"] * blk.inlet.flow_mol[0]
        return blk.rate_reaction_extent[0, "Ca_removal"] == smooth_min(a, b, eps=1e-20)

    model.fs.stoich_softening_reactor_unit.calcium_extent_con = Constraint(
        rule=_ca_extent_cons
    )

    def _mg_extent_cons(blk):
        a = (
            blk.inlet.mole_frac_comp[0, "Ca(OH)2"]
            * (1 - blk.frac_excess_lime)
            * blk.inlet.flow_mol[0]
        )
        b = blk.inlet.mole_frac_comp[0, "Mg(HCO3)2"] * blk.inlet.flow_mol[0]
        return blk.rate_reaction_extent[0, "Mg_removal"] == smooth_min(
            (a - blk.rate_reaction_extent[0, "Ca_removal"]) / 2, b, eps=1e-20
        )

    model.fs.stoich_softening_reactor_unit.magnesium_extent_con = Constraint(
        rule=_mg_extent_cons
    )


def set_stoich_softening_separator_inlets(
    model,
    residual_lime_mg_per_L=2,
    inlet_water_density_kg_per_L=1,
    inlet_temperature_K=298,
    inlet_pressure_Pa=101325,
    inlet_flow_mol_per_s=10,
    inlet_carbonate_hardness_mg_per_L=60,
    hardness_fraction_to_Ca=0.5,
    inlet_salinity_psu=35,
    inlet_excess_chlorine_ppm=35,
    inlet_sulfate_ppm=2000,
    inlet_solids_mg_per_L=140,
):
    model.fs.stoich_softening_separator_unit.inlet.pressure[0].set_value(
        inlet_pressure_Pa
    )
    model.fs.stoich_softening_separator_unit.inlet.temperature[0].set_value(
        inlet_temperature_K
    )
    model.fs.stoich_softening_separator_unit.inlet.flow_mol[0].set_value(
        inlet_flow_mol_per_s
    )

    zero_out_non_H2O_molefractions(model.fs.stoich_softening_separator_unit.inlet)
    # Calculate molefractions for Ca(HCO3)2 and Mg(HCO3)2
    total_molar_density = inlet_water_density_kg_per_L / 18 * 1000  # mol/L
    if hardness_fraction_to_Ca > 1:
        hardness_fraction_to_Ca = 1
    if hardness_fraction_to_Ca < 0:
        hardness_fraction_to_Ca = 0
    x_Ca = (
        inlet_carbonate_hardness_mg_per_L / 50000 / 2 / total_molar_density
    ) * hardness_fraction_to_Ca
    x_Mg = (inlet_carbonate_hardness_mg_per_L / 50000 / 2 / total_molar_density) * (
        1 - hardness_fraction_to_Ca
    )
    model.fs.stoich_softening_separator_unit.inlet.mole_frac_comp[
        0, "Ca(HCO3)2"
    ].set_value(x_Ca)
    model.fs.stoich_softening_separator_unit.inlet.mole_frac_comp[
        0, "Mg(HCO3)2"
    ].set_value(x_Mg)

    x_so4 = (
        inlet_sulfate_ppm * inlet_water_density_kg_per_L / 96060 / total_molar_density
    )
    model.fs.stoich_softening_separator_unit.inlet.mole_frac_comp[
        0, "SO4_2-"
    ].set_value(x_so4)

    x_cl = (
        inlet_excess_chlorine_ppm
        * inlet_water_density_kg_per_L
        / 96060
        / total_molar_density
    )
    model.fs.stoich_softening_separator_unit.inlet.mole_frac_comp[0, "Cl_-"].set_value(
        x_cl
    )

    total_salt = (
        value(
            model.fs.stoich_softening_separator_unit.inlet.mole_frac_comp[
                0, "Ca(HCO3)2"
            ]
        )
        * total_molar_density
        * 101
    )
    total_salt += (
        value(
            model.fs.stoich_softening_separator_unit.inlet.mole_frac_comp[
                0, "Mg(HCO3)2"
            ]
        )
        * total_molar_density
        * 85.31
    )
    psu_from_hardness = total_salt / (total_molar_density * 18) * 1000
    if psu_from_hardness < inlet_salinity_psu:
        psu_from_nacl = inlet_salinity_psu - psu_from_hardness
        x_NaCl = (
            psu_from_nacl
            * (total_molar_density * 18)
            / 1000
            / total_molar_density
            / 58.44
        )
        model.fs.stoich_softening_separator_unit.inlet.mole_frac_comp[
            0, "NaCl"
        ].set_value(x_NaCl)

    x_lime = residual_lime_mg_per_L / 1000 / 74.093 / total_molar_density
    model.fs.stoich_softening_separator_unit.inlet.mole_frac_comp[
        0, "Ca(OH)2"
    ].set_value(x_lime)

    x_CaCO3 = (
        inlet_solids_mg_per_L / 50000 / 2 / total_molar_density
    ) * hardness_fraction_to_Ca
    x_MgOH = (inlet_solids_mg_per_L / 50000 / 2 / total_molar_density) * (
        1 - hardness_fraction_to_Ca
    )
    model.fs.stoich_softening_separator_unit.inlet.mole_frac_comp[0, "CaCO3"].set_value(
        x_CaCO3
    )
    model.fs.stoich_softening_separator_unit.inlet.mole_frac_comp[
        0, "Mg(OH)2"
    ].set_value(x_MgOH)

    set_H2O_molefraction(model.fs.stoich_softening_separator_unit.inlet)


def set_stoich_softening_separator_split_frac(model, solids_removal_frac=0.99):
    model.fs.stoich_softening_separator_unit.split_fraction[
        0, "outlet_stream", "H2O"
    ].set_value(solids_removal_frac)
    model.fs.stoich_softening_separator_unit.split_fraction[
        0, "outlet_stream", "NaCl"
    ].set_value(solids_removal_frac)
    model.fs.stoich_softening_separator_unit.split_fraction[
        0, "outlet_stream", "Ca(OH)2"
    ].set_value((1 - solids_removal_frac))
    model.fs.stoich_softening_separator_unit.split_fraction[
        0, "outlet_stream", "Mg(OH)2"
    ].set_value((1 - solids_removal_frac))
    model.fs.stoich_softening_separator_unit.split_fraction[
        0, "outlet_stream", "Ca(HCO3)2"
    ].set_value(solids_removal_frac)
    model.fs.stoich_softening_separator_unit.split_fraction[
        0, "outlet_stream", "Mg(HCO3)2"
    ].set_value(solids_removal_frac)
    model.fs.stoich_softening_separator_unit.split_fraction[
        0, "outlet_stream", "CaCO3"
    ].set_value((1 - solids_removal_frac))
    model.fs.stoich_softening_separator_unit.split_fraction[
        0, "outlet_stream", "SO4_2-"
    ].set_value(solids_removal_frac)
    model.fs.stoich_softening_separator_unit.split_fraction[
        0, "outlet_stream", "Cl_-"
    ].set_value(solids_removal_frac)


def fix_stoich_softening_mixer_inlet_stream(model):
    model.fs.stoich_softening_mixer_unit.inlet_stream.flow_mol[0].fix()
    model.fs.stoich_softening_mixer_unit.inlet_stream.pressure[0].fix()
    model.fs.stoich_softening_mixer_unit.inlet_stream.temperature[0].fix()
    fix_all_molefractions(model.fs.stoich_softening_mixer_unit.inlet_stream)


def fix_stoich_softening_mixer_lime_stream(model):
    model.fs.stoich_softening_mixer_unit.lime_stream.flow_mol[0].fix()
    model.fs.stoich_softening_mixer_unit.lime_stream.pressure[0].fix()
    model.fs.stoich_softening_mixer_unit.lime_stream.temperature[0].fix()
    fix_all_molefractions(model.fs.stoich_softening_mixer_unit.lime_stream)


def unfix_stoich_softening_mixer_inlet_stream(model):
    model.fs.stoich_softening_mixer_unit.inlet_stream.flow_mol[0].unfix()
    model.fs.stoich_softening_mixer_unit.inlet_stream.pressure[0].unfix()
    model.fs.stoich_softening_mixer_unit.inlet_stream.temperature[0].unfix()
    unfix_all_molefractions(model.fs.stoich_softening_mixer_unit.inlet_stream)


def unfix_stoich_softening_mixer_lime_stream(model):
    model.fs.stoich_softening_mixer_unit.lime_stream.flow_mol[0].unfix()
    model.fs.stoich_softening_mixer_unit.lime_stream.pressure[0].unfix()
    model.fs.stoich_softening_mixer_unit.lime_stream.temperature[0].unfix()
    unfix_all_molefractions(model.fs.ideal_naocl_mixer_unit.lime_stream)


def fix_stoich_softening_reactor_inlets(model):
    model.fs.stoich_softening_reactor_unit.inlet.flow_mol[0].fix()
    model.fs.stoich_softening_reactor_unit.inlet.pressure[0].fix()
    model.fs.stoich_softening_reactor_unit.inlet.temperature[0].fix()
    fix_all_molefractions(model.fs.stoich_softening_reactor_unit.inlet)


def unfix_stoich_softening_reactor_inlets(model):
    model.fs.stoich_softening_reactor_unit.inlet.flow_mol[0].unfix()
    model.fs.stoich_softening_reactor_unit.inlet.pressure[0].unfix()
    model.fs.stoich_softening_reactor_unit.inlet.temperature[0].unfix()
    unfix_all_molefractions(model.fs.stoich_softening_reactor_unit.inlet)


def unfix_stoich_softening_reactor_extents(model):
    model.fs.stoich_softening_reactor_unit.rate_reaction_extent[0, "Ca_removal"].unfix()
    model.fs.stoich_softening_reactor_unit.rate_reaction_extent[0, "Mg_removal"].unfix()


def fix_stoich_softening_separator_split_frac(model):
    model.fs.stoich_softening_separator_unit.split_fraction[
        0, "outlet_stream", "H2O"
    ].fix()
    model.fs.stoich_softening_separator_unit.split_fraction[
        0, "outlet_stream", "NaCl"
    ].fix()
    model.fs.stoich_softening_separator_unit.split_fraction[
        0, "outlet_stream", "Ca(OH)2"
    ].fix()
    model.fs.stoich_softening_separator_unit.split_fraction[
        0, "outlet_stream", "Mg(OH)2"
    ].fix()
    model.fs.stoich_softening_separator_unit.split_fraction[
        0, "outlet_stream", "Ca(HCO3)2"
    ].fix()
    model.fs.stoich_softening_separator_unit.split_fraction[
        0, "outlet_stream", "Mg(HCO3)2"
    ].fix()
    model.fs.stoich_softening_separator_unit.split_fraction[
        0, "outlet_stream", "CaCO3"
    ].fix()
    model.fs.stoich_softening_separator_unit.split_fraction[
        0, "outlet_stream", "SO4_2-"
    ].fix()
    model.fs.stoich_softening_separator_unit.split_fraction[
        0, "outlet_stream", "Cl_-"
    ].fix()


def fix_stoich_softening_separator_inlets(model):
    model.fs.stoich_softening_separator_unit.inlet.pressure.fix()
    model.fs.stoich_softening_separator_unit.inlet.temperature.fix()
    model.fs.stoich_softening_separator_unit.inlet.flow_mol.fix()

    fix_all_molefractions(model.fs.stoich_softening_separator_unit.inlet)


def unfix_stoich_softening_separator_split_frac(model):
    model.fs.stoich_softening_separator_unit.split_fraction[
        0, "outlet_stream", "H2O"
    ].unfix()
    model.fs.stoich_softening_separator_unit.split_fraction[
        0, "outlet_stream", "NaCl"
    ].unfix()
    model.fs.stoich_softening_separator_unit.split_fraction[
        0, "outlet_stream", "Ca(OH)2"
    ].unfix()
    model.fs.stoich_softening_separator_unit.split_fraction[
        0, "outlet_stream", "Mg(OH)2"
    ].unfix()
    model.fs.stoich_softening_separator_unit.split_fraction[
        0, "outlet_stream", "Ca(HCO3)2"
    ].unfix()
    model.fs.stoich_softening_separator_unit.split_fraction[
        0, "outlet_stream", "Mg(HCO3)2"
    ].unfix()
    model.fs.stoich_softening_separator_unit.split_fraction[
        0, "outlet_stream", "CaCO3"
    ].unfix()
    model.fs.stoich_softening_separator_unit.split_fraction[
        0, "outlet_stream", "SO4_2-"
    ].unfix()
    model.fs.stoich_softening_separator_unit.split_fraction[
        0, "outlet_stream", "Cl_-"
    ].unfix()


def unfix_stoich_softening_separator_inlets(model):
    model.fs.stoich_softening_separator_unit.inlet.pressure.fix()
    model.fs.stoich_softening_separator_unit.inlet.temperature.fix()
    model.fs.stoich_softening_separator_unit.inlet.flow_mol.fix()

    unfix_all_molefractions(model.fs.stoich_softening_separator_unit.inlet)


def initialize_stoich_softening_mixer(unit, debug_out=False):
    was_fixed = False
    if not unit.lime_stream.flow_mol[0].is_fixed():
        unit.lime_stream.flow_mol[0].fix()
        was_fixed = True
    if debug_out:
        unit.initialize(optarg=solver.options, outlvl=idaeslog.DEBUG)
    else:
        unit.initialize(optarg=solver.options)
    if was_fixed:
        unit.lime_stream.flow_mol[0].unfix()


def initialize_stoich_softening_reactor(unit, debug_out=False):
    if debug_out:
        unit.initialize(optarg=solver.options, outlvl=idaeslog.DEBUG)
    else:
        unit.initialize(optarg=solver.options)


def initialize_stoich_softening_separator(unit, debug_out=False):
    if debug_out:
        unit.initialize(optarg=solver.options, outlvl=idaeslog.DEBUG)
    else:
        unit.initialize(optarg=solver.options)


def setup_block_to_solve_lime_dosing_rate(model, target_hardness_mg_per_L=50):
    model.fs.stoich_softening_mixer_unit.lime_stream.flow_mol[0].unfix()
    model.fs.stoich_softening_separator_unit.hardness.fix(target_hardness_mg_per_L)


def display_results_of_stoich_softening_mixer(unit):
    print()
    print("=========== Stoich Softening Mixer Results ============")
    print("Outlet Temperature:       \t" + str(unit.outlet.temperature[0].value))
    print("Outlet Pressure:          \t" + str(unit.outlet.pressure[0].value))
    print("Outlet FlowMole:          \t" + str(unit.outlet.flow_mol[0].value))
    print()
    total_molar_density = 55.6
    total_salt = (
        value(unit.outlet.mole_frac_comp[0, "Ca(OH)2"]) * total_molar_density * 74.093
    )
    total_salt += (
        value(unit.outlet.mole_frac_comp[0, "Ca(HCO3)2"]) * total_molar_density * 101
    )
    total_salt += (
        value(unit.outlet.mole_frac_comp[0, "Mg(HCO3)2"]) * total_molar_density * 85.31
    )
    total_salt += (
        value(unit.outlet.mole_frac_comp[0, "CaCO3"]) * total_molar_density * 100
    )
    total_salt += (
        value(unit.outlet.mole_frac_comp[0, "Mg(OH)2"]) * total_molar_density * 58.32
    )
    total_salt += (
        value(unit.outlet.mole_frac_comp[0, "NaCl"]) * total_molar_density * 58.44
    )
    psu = total_salt / (total_molar_density * 18) * 1000
    print("STP Salinity (PSU):           \t" + str(psu))
    total_hardness = (
        50000
        * 2
        * value(unit.inlet_stream.mole_frac_comp[0, "Ca(OH)2"])
        * total_molar_density
    )
    total_hardness += (
        50000
        * 2
        * value(unit.inlet_stream.mole_frac_comp[0, "Ca(HCO3)2"])
        * total_molar_density
    )
    total_hardness += (
        50000
        * 2
        * value(unit.inlet_stream.mole_frac_comp[0, "Mg(HCO3)2"])
        * total_molar_density
    )
    total_hardness += (
        50000
        * 2
        * value(unit.inlet_stream.mole_frac_comp[0, "CaCO3"])
        * total_molar_density
    )
    total_hardness += (
        50000
        * 2
        * value(unit.inlet_stream.mole_frac_comp[0, "Mg(OH)2"])
        * total_molar_density
    )
    print("Inlet Hardness (mg/L):   \t" + str(total_hardness))
    print("Lime Dosing Rate (kg/s): \t" + str(unit.dosing_rate.value))
    print(
        "Lime Dosage (mg/L):      \t"
        + str(
            value(unit.outlet.mole_frac_comp[0, "Ca(OH)2"])
            * total_molar_density
            * 74.093
            * 1000
        )
    )
    print("------------------------------------------------------")
    print()


def display_results_of_stoich_softening_reactor(unit):
    print()
    print("=========== Stoich Softening Reactor Results ============")
    print("Outlet Temperature:       \t" + str(unit.outlet.temperature[0].value))
    print("Outlet Pressure:          \t" + str(unit.outlet.pressure[0].value))
    print("Outlet FlowMole:          \t" + str(unit.outlet.flow_mol[0].value))
    print()
    total_molar_density = 55.6
    total_salt = (
        value(unit.outlet.mole_frac_comp[0, "Ca(OH)2"]) * total_molar_density * 74.093
    )
    total_salt += (
        value(unit.outlet.mole_frac_comp[0, "Ca(HCO3)2"]) * total_molar_density * 101
    )
    total_salt += (
        value(unit.outlet.mole_frac_comp[0, "Mg(HCO3)2"]) * total_molar_density * 85.31
    )
    total_salt += (
        value(unit.outlet.mole_frac_comp[0, "CaCO3"]) * total_molar_density * 100
    )
    total_salt += (
        value(unit.outlet.mole_frac_comp[0, "Mg(OH)2"]) * total_molar_density * 58.32
    )
    total_salt += (
        value(unit.outlet.mole_frac_comp[0, "NaCl"]) * total_molar_density * 58.44
    )
    psu = total_salt / (total_molar_density * 18) * 1000
    print("STP Salinity (PSU):           \t" + str(psu))

    total_hardness = (
        50000
        * 2
        * value(unit.inlet.mole_frac_comp[0, "Ca(HCO3)2"])
        * total_molar_density
    )
    total_hardness += (
        50000
        * 2
        * value(unit.inlet.mole_frac_comp[0, "Mg(HCO3)2"])
        * total_molar_density
    )
    total_hardness += (
        50000 * 2 * value(unit.inlet.mole_frac_comp[0, "CaCO3"]) * total_molar_density
    )
    total_hardness += (
        50000 * 2 * value(unit.inlet.mole_frac_comp[0, "Mg(OH)2"]) * total_molar_density
    )
    print(
        "Inlet Hardness (mg/L):   \t"
        + str(total_hardness)
        + "\t(not including lime added)"
    )
    res_hardness = (
        50000
        * 2
        * value(unit.outlet.mole_frac_comp[0, "Ca(HCO3)2"])
        * total_molar_density
    )
    res_hardness += (
        50000
        * 2
        * value(unit.outlet.mole_frac_comp[0, "Mg(HCO3)2"])
        * total_molar_density
    )
    print(
        "Residual Hardness(mg/L):\t"
        + str(res_hardness)
        + "\t(only includes bicarbonate species)"
    )
    print(
        "Lime Dosage (mg/L):      \t"
        + str(
            value(unit.inlet.mole_frac_comp[0, "Ca(OH)2"])
            * total_molar_density
            * 74.093
            * 1000
        )
    )
    print(
        "Lime Residual (mg/L):    \t"
        + str(
            value(unit.outlet.mole_frac_comp[0, "Ca(OH)2"])
            * total_molar_density
            * 74.093
            * 1000
        )
    )
    print("-------------------------------------------------------")
    print()


def display_results_of_stoich_softening_separator(unit):
    print()
    print("=========== Stoich Softening Separator Results ============")
    print("Outlet Temperature:       \t" + str(unit.outlet_stream.temperature[0].value))
    print("Outlet Pressure:          \t" + str(unit.outlet_stream.pressure[0].value))
    print("Outlet FlowMole:          \t" + str(unit.outlet_stream.flow_mol[0].value))
    print()
    total_molar_density = value(unit.mixed_state[0.0].dens_mol_phase["Liq"]) / 1000
    total_salt = (
        value(unit.outlet_stream.mole_frac_comp[0, "Ca(OH)2"])
        * total_molar_density
        * 74.093
    )
    total_salt += (
        value(unit.outlet_stream.mole_frac_comp[0, "Ca(HCO3)2"])
        * total_molar_density
        * 101
    )
    total_salt += (
        value(unit.outlet_stream.mole_frac_comp[0, "Mg(HCO3)2"])
        * total_molar_density
        * 85.31
    )
    total_salt += (
        value(unit.outlet_stream.mole_frac_comp[0, "CaCO3"]) * total_molar_density * 100
    )
    total_salt += (
        value(unit.outlet_stream.mole_frac_comp[0, "Mg(OH)2"])
        * total_molar_density
        * 58.32
    )
    total_salt += (
        value(unit.outlet_stream.mole_frac_comp[0, "NaCl"])
        * total_molar_density
        * 58.44
    )
    psu = total_salt / (total_molar_density * 18) * 1000
    print("Salinity (PSU):           \t" + str(psu))
    total_hardness = (
        50000 * 2 * value(unit.inlet.mole_frac_comp[0, "Ca(OH)2"]) * total_molar_density
    )
    total_hardness += (
        50000
        * 2
        * value(unit.inlet.mole_frac_comp[0, "Ca(HCO3)2"])
        * total_molar_density
    )
    total_hardness += (
        50000
        * 2
        * value(unit.inlet.mole_frac_comp[0, "Mg(HCO3)2"])
        * total_molar_density
    )
    total_hardness += (
        50000 * 2 * value(unit.inlet.mole_frac_comp[0, "CaCO3"]) * total_molar_density
    )
    total_hardness += (
        50000 * 2 * value(unit.inlet.mole_frac_comp[0, "Mg(OH)2"]) * total_molar_density
    )
    print("Inlet Hardness (mg/L):  \t" + str(total_hardness))
    print("Residual Hardness(mg/L):\t" + str(unit.hardness.value))
    print("-------------------------------------------------------")
    print()


def build_stoich_softening_block(
    model,
    include_feed=False,
    expand_arcs=False,
    frac_excess_lime=0.01,
    frac_used_for_Ca_removal=0.99,
    solids_removal_frac=0.99,
):
    # Add properties to model
    build_stoich_softening_prop(model)

    # Add mixer to the model
    build_stoich_softening_mixer_unit(model)

    # Add reactor to the model
    build_stoich_softening_reactor_unit(
        model,
        frac_excess_lime=frac_excess_lime,
        frac_used_for_Ca_removal=frac_used_for_Ca_removal,
    )

    # Add separator to the model
    build_stoich_softening_separator_unit(
        model, solids_removal_frac=solids_removal_frac
    )

    if include_feed == True:
        # add the feed to the model
        # # TODO: build and test feed unit
        raise RuntimeError("Feed not available yet...")

    # Connect the mixer to the reactor unit with arcs
    model.fs.stoich_softening_arc_mixer_to_reactor = Arc(
        source=model.fs.stoich_softening_mixer_unit.outlet,
        destination=model.fs.stoich_softening_reactor_unit.inlet,
    )

    # Connect the mixer to the reactor unit with arcs
    model.fs.stoich_softening_arc_reactor_to_separator = Arc(
        source=model.fs.stoich_softening_reactor_unit.outlet,
        destination=model.fs.stoich_softening_separator_unit.inlet,
    )

    if include_feed == True:
        # add Arc from feed to mixer
        # # TODO: build and test feed unit
        raise RuntimeError("Feed not available yet...")

    if expand_arcs == True:
        TransformationFactory("network.expand_arcs").apply_to(model)


def run_stoich_softening_mixer_example():
    model = ConcreteModel()
    model.fs = FlowsheetBlock(dynamic=False)

    # Add properties to model
    build_stoich_softening_prop(model)

    # add the mixer
    build_stoich_softening_mixer_unit(model)

    # set model values
    set_stoich_softening_mixer_inlets(model)

    # fix inlets for testing
    fix_stoich_softening_mixer_inlet_stream(model)
    fix_stoich_softening_mixer_lime_stream(model)

    check_dof(model)

    # initializer mixer
    initialize_stoich_softening_mixer(
        model.fs.stoich_softening_mixer_unit, debug_out=False
    )

    # solve with user scaling
    solve_block(model, tee=True)

    model.fs.stoich_softening_mixer_unit.outlet.mole_frac_comp.pprint()
    model.fs.stoich_softening_mixer_unit.dosing_rate.pprint()

    display_results_of_stoich_softening_mixer(model.fs.stoich_softening_mixer_unit)

    return model


def run_stoich_softening_reactor_example():
    model = ConcreteModel()
    model.fs = FlowsheetBlock(dynamic=False)

    # Add properties to model
    build_stoich_softening_prop(model)

    # add the reactor
    build_stoich_softening_reactor_unit(
        model, frac_excess_lime=0.01, frac_used_for_Ca_removal=0.5
    )

    # set some values
    set_stoich_softening_reactor_inlets(
        model,
        dosage_of_lime_mg_per_L=140,
        inlet_water_density_kg_per_L=1,
        inlet_temperature_K=298,
        inlet_pressure_Pa=101325,
        inlet_flow_mol_per_s=10,
        inlet_total_hardness_mg_per_L=200,
        hardness_fraction_to_Ca=0.5,
        inlet_salinity_psu=35,
        inlet_sulfate_ppm=2000,
    )

    # fix inlets for testing
    fix_stoich_softening_reactor_inlets(model)

    check_dof(model)

    # initializer reactor
    initialize_stoich_softening_reactor(
        model.fs.stoich_softening_reactor_unit, debug_out=False
    )

    # solve with user scaling
    solve_block(model, tee=True)

    display_results_of_stoich_softening_reactor(model.fs.stoich_softening_reactor_unit)
    model.fs.stoich_softening_reactor_unit.outlet.mole_frac_comp.pprint()

    return model


def run_stoich_softening_separator_example():
    model = ConcreteModel()
    model.fs = FlowsheetBlock(dynamic=False)

    # Add properties to model
    build_stoich_softening_prop(model)

    # add the separator
    build_stoich_softening_separator_unit(model)

    # set up values for separator
    set_stoich_softening_separator_inlets(model)

    # fix the values to close the system of equations
    fix_stoich_softening_separator_inlets(model)

    check_dof(model)

    # initializer separator
    initialize_stoich_softening_separator(
        model.fs.stoich_softening_separator_unit, debug_out=False
    )

    # solve with user scaling
    solve_block(model, tee=True)

    display_results_of_stoich_softening_separator(
        model.fs.stoich_softening_separator_unit
    )

    return model


def run_softening_block_example(include_feed=False, fix_hardness=False):
    model = ConcreteModel()
    model.fs = FlowsheetBlock(dynamic=False)

    build_stoich_softening_block(model, include_feed=include_feed, expand_arcs=True)

    if include_feed == True:
        # set feed inlets
        # # TODO: build and test feed unit
        raise RuntimeError("Feed not available yet...")

    # set mixer inlets values (mostly done for the lime dosage)
    set_stoich_softening_mixer_inlets(
        model,
        dosing_rate_of_lime_mg_per_s=25,
        inlet_water_density_kg_per_L=1,
        inlet_temperature_K=298,
        inlet_pressure_Pa=101325,
        inlet_flow_mol_per_s=10,
        inlet_total_hardness_mg_per_L=200,
        hardness_fraction_to_Ca=0.5,
        inlet_salinity_psu=35,
        inlet_excess_chlorine_ppm=35,
        inlet_sulfate_ppm=2000,
    )

    if include_feed == True:
        # fix feed inlets
        raise RuntimeError("Feed not available yet...")
    else:
        # fix inlets at the mixer for testing
        fix_stoich_softening_mixer_inlet_stream(model)

    # test the stoich_softening_mixer  costing
    model.fs.treated_flow_vol = Expression(expr=0.0007 * pyunits.m**3 / pyunits.s)
    costing.build_costing(model)

    # Fix the amount of lime added for simulation
    fix_stoich_softening_mixer_lime_stream(model)

    check_dof(model)

    # ----------- Scale each block --------------
    if include_feed == True:
        # scale the feed and propogate the state
        # # TODO: build and test feed unit
        raise RuntimeError("Feed not available yet...")

    # --------- Initialize: Works best to manually initialize with state propogation
    if include_feed == True:
        # initialize the feed and propogate the state
        # # TODO: build and test feed unit
        raise RuntimeError("Feed not available yet...")

    # initializer mixer
    initialize_stoich_softening_mixer(
        model.fs.stoich_softening_mixer_unit, debug_out=False
    )
    # initializer reactor
    propagate_state(model.fs.stoich_softening_arc_mixer_to_reactor)
    initialize_stoich_softening_reactor(
        model.fs.stoich_softening_reactor_unit, debug_out=False
    )
    # initializer separator
    propagate_state(model.fs.stoich_softening_arc_reactor_to_separator)
    initialize_stoich_softening_separator(
        model.fs.stoich_softening_separator_unit, debug_out=False
    )

    # Solve the flowsheet
    if fix_hardness == True:
        setup_block_to_solve_lime_dosing_rate(model, target_hardness_mg_per_L=50)

    model.fs.costing.initialize()

    solve_block(model, tee=True)

    # display results
    display_results_of_stoich_softening_mixer(model.fs.stoich_softening_mixer_unit)
    display_results_of_stoich_softening_reactor(model.fs.stoich_softening_reactor_unit)
    display_results_of_stoich_softening_separator(
        model.fs.stoich_softening_separator_unit
    )

    return model



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
Flowsheet example full Water Resource Recovery Facility 
(WRRF; a.k.a., wastewater treatment plant) with ASM1 and ADM1.

The flowsheet follows the same formulation as benchmark simulation model no.2 (BSM2)
but comprises different specifications for default values than BSM2.

"""
__author__ = "Alejandro Garciadiego, Xinhong Liu, Adam Atia"

import pyomo.environ as pyo

from pyomo.network import Arc, SequentialDecomposition
from watertap.unit_models.anaerobic_digestor import AD
from watertap.unit_models.thickener import Thickener
from watertap.unit_models.dewatering import DewateringUnit
from watertap.unit_models.cstr import CSTR
from watertap.unit_models.clarifier import Clarifier

from watertap.unit_models.translators.translator_asm1_adm1 import Translator_ASM1_ADM1
from watertap.unit_models.translators.translator_adm1_asm1 import Translator_ADM1_ASM1

import idaes.logger as idaeslog
from idaes.core.solvers import get_solver
import idaes.core.util.scaling as iscale

from watertap.property_models.anaerobic_digestion.adm1_properties import (
    ADM1ParameterBlock,
)
from watertap.property_models.anaerobic_digestion.adm1_reactions import (
    ADM1ReactionParameterBlock,
)
from idaes.models.unit_models.separator import SplittingType
from watertap.property_models.anaerobic_digestion.adm1_properties_vapor import (
    ADM1_vaporParameterBlock,
)

from idaes.core import FlowsheetBlock, UnitModelCostingBlock
from idaes.models.unit_models import (
    Feed,
    Mixer,
    Separator,
    PressureChanger,
    Product,
)

from watertap.unit_models.cstr_injection import CSTR_Injection
from watertap.property_models.activated_sludge.asm1_properties import ASM1ParameterBlock
from watertap.property_models.activated_sludge.asm1_reactions import (
    ASM1ReactionParameterBlock,
)
from watertap.core.util.initialization import assert_degrees_of_freedom
from watertap.costing import WaterTAPCosting
from watertap.costing.unit_models.clarifier import (
    cost_circular_clarifier,
    cost_primary_clarifier,
)
from pyomo.util.check_units import assert_units_consistent


def main():
    m = build()
    set_operating_conditions(m)
    assert_degrees_of_freedom(m, 0)
    assert_units_consistent(m)

    initialize_system(m)

    results = solve(m)

    add_costing(m)
    m.fs.costing.initialize()
    assert_degrees_of_freedom(m, 0)

    results = solve(m)
    pyo.assert_optimal_termination(results)
    display_results(m)
    display_costing(m)

    return m, results


def build():
    m = pyo.ConcreteModel()

    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.props_ASM1 = ASM1ParameterBlock()
    m.fs.props_ADM1 = ADM1ParameterBlock()
    m.fs.props_vap = ADM1_vaporParameterBlock()
    m.fs.ADM1_rxn_props = ADM1ReactionParameterBlock(property_package=m.fs.props_ADM1)
    m.fs.ASM1_rxn_props = ASM1ReactionParameterBlock(property_package=m.fs.props_ASM1)
    # Feed water stream
    m.fs.FeedWater = Feed(property_package=m.fs.props_ASM1)

    # ==========================================================================
    # Activated Sludge Process
    # ==========================================================================
    # Mixer for inlet water and recycled sludge
    m.fs.MX1 = Mixer(
        property_package=m.fs.props_ASM1, inlet_list=["feed_water", "recycle"]
    )
    # First reactor (anoxic) - standard CSTR
    m.fs.R1 = CSTR(
        property_package=m.fs.props_ASM1, reaction_package=m.fs.ASM1_rxn_props
    )
    # Second reactor (anoxic) - standard CSTR
    m.fs.R2 = CSTR(
        property_package=m.fs.props_ASM1, reaction_package=m.fs.ASM1_rxn_props
    )
    # Third reactor (aerobic) - CSTR with injection
    m.fs.R3 = CSTR_Injection(
        property_package=m.fs.props_ASM1, reaction_package=m.fs.ASM1_rxn_props
    )
    # Fourth reactor (aerobic) - CSTR with injection
    m.fs.R4 = CSTR_Injection(
        property_package=m.fs.props_ASM1, reaction_package=m.fs.ASM1_rxn_props
    )
    # Fifth reactor (aerobic) - CSTR with injection
    m.fs.R5 = CSTR_Injection(
        property_package=m.fs.props_ASM1, reaction_package=m.fs.ASM1_rxn_props
    )
    m.fs.SP5 = Separator(
        property_package=m.fs.props_ASM1, outlet_list=["underflow", "overflow"]
    )
    # Clarifier
    # TODO: Replace with more detailed model when available
    m.fs.CL1 = Clarifier(
        property_package=m.fs.props_ASM1,
        outlet_list=["underflow", "effluent"],
        split_basis=SplittingType.componentFlow,
    )
    # Sludge purge splitter
    m.fs.SP6 = Separator(
        property_package=m.fs.props_ASM1,
        outlet_list=["recycle", "waste"],
        split_basis=SplittingType.totalFlow,
    )
    # Mixing sludge recycle and R5 underflow
    m.fs.MX6 = Mixer(
        property_package=m.fs.props_ASM1, inlet_list=["clarifier", "reactor"]
    )
    # Product Blocks
    m.fs.Treated = Product(property_package=m.fs.props_ASM1)
    # Recycle pressure changer - use a simple isothermal unit for now
    m.fs.P1 = PressureChanger(property_package=m.fs.props_ASM1)

    # Link units
    m.fs.stream2 = Arc(source=m.fs.MX1.outlet, destination=m.fs.R1.inlet)
    m.fs.stream3 = Arc(source=m.fs.R1.outlet, destination=m.fs.R2.inlet)
    m.fs.stream4 = Arc(source=m.fs.R2.outlet, destination=m.fs.R3.inlet)
    m.fs.stream5 = Arc(source=m.fs.R3.outlet, destination=m.fs.R4.inlet)
    m.fs.stream6 = Arc(source=m.fs.R4.outlet, destination=m.fs.R5.inlet)
    m.fs.stream7 = Arc(source=m.fs.R5.outlet, destination=m.fs.SP5.inlet)
    m.fs.stream8 = Arc(source=m.fs.SP5.overflow, destination=m.fs.CL1.inlet)
    m.fs.stream9 = Arc(source=m.fs.SP5.underflow, destination=m.fs.MX6.reactor)
    m.fs.stream10 = Arc(source=m.fs.CL1.effluent, destination=m.fs.Treated.inlet)
    m.fs.stream11 = Arc(source=m.fs.CL1.underflow, destination=m.fs.SP6.inlet)
    m.fs.stream13 = Arc(source=m.fs.SP6.recycle, destination=m.fs.MX6.clarifier)
    m.fs.stream14 = Arc(source=m.fs.MX6.outlet, destination=m.fs.P1.inlet)
    m.fs.stream15 = Arc(source=m.fs.P1.outlet, destination=m.fs.MX1.recycle)
    pyo.TransformationFactory("network.expand_arcs").apply_to(m)

    # Oxygen concentration in reactors 3 and 4 is governed by mass transfer
    # Add additional parameter and constraints
    m.fs.R3.KLa = pyo.Var(
        initialize=7.6,
        units=pyo.units.hour**-1,
        doc="Lumped mass transfer coefficient for oxygen",
    )
    m.fs.R4.KLa = pyo.Var(
        initialize=5.7,
        units=pyo.units.hour**-1,
        doc="Lumped mass transfer coefficient for oxygen",
    )
    m.fs.S_O_eq = pyo.Param(
        default=8e-3,
        units=pyo.units.kg / pyo.units.m**3,
        mutable=True,
        doc="Dissolved oxygen concentration at equilibrium",
    )

    @m.fs.R3.Constraint(m.fs.time, doc="Mass transfer constraint for R3")
    def mass_transfer_R3(self, t):
        return pyo.units.convert(
            m.fs.R3.injection[t, "Liq", "S_O"], to_units=pyo.units.kg / pyo.units.hour
        ) == (
            m.fs.R3.KLa
            * m.fs.R3.volume[t]
            * (m.fs.S_O_eq - m.fs.R3.outlet.conc_mass_comp[t, "S_O"])
        )

    @m.fs.R4.Constraint(m.fs.time, doc="Mass transfer constraint for R4")
    def mass_transfer_R4(self, t):
        return pyo.units.convert(
            m.fs.R4.injection[t, "Liq", "S_O"], to_units=pyo.units.kg / pyo.units.hour
        ) == (
            m.fs.R4.KLa
            * m.fs.R4.volume[t]
            * (m.fs.S_O_eq - m.fs.R4.outlet.conc_mass_comp[t, "S_O"])
        )

    # ======================================================================
    # Anaerobic digester section
    m.fs.asm_adm = Translator_ASM1_ADM1(
        inlet_property_package=m.fs.props_ASM1,
        outlet_property_package=m.fs.props_ADM1,
        reaction_package=m.fs.ADM1_rxn_props,
        has_phase_equilibrium=False,
        outlet_state_defined=True,
    )

    m.fs.RADM = AD(
        liquid_property_package=m.fs.props_ADM1,
        vapor_property_package=m.fs.props_vap,
        reaction_package=m.fs.ADM1_rxn_props,
        has_heat_transfer=True,
        has_pressure_change=False,
    )

    m.fs.adm_asm = Translator_ADM1_ASM1(
        inlet_property_package=m.fs.props_ADM1,
        outlet_property_package=m.fs.props_ASM1,
        reaction_package=m.fs.ADM1_rxn_props,
        has_phase_equilibrium=False,
        outlet_state_defined=True,
    )

    # ====================================================================
    # Primary Clarifier
    m.fs.CL = Clarifier(
        property_package=m.fs.props_ASM1,
        outlet_list=["underflow", "effluent"],
        split_basis=SplittingType.componentFlow,
    )

    # Thickener
    m.fs.TU = Thickener(property_package=m.fs.props_ASM1)
    # Dewaterer
    m.fs.DU = DewateringUnit(property_package=m.fs.props_ASM1)

    m.fs.MX2 = Mixer(
        property_package=m.fs.props_ASM1, inlet_list=["feed_water1", "recycle1"]
    )
    m.fs.MX3 = Mixer(
        property_package=m.fs.props_ASM1, inlet_list=["feed_water2", "recycle2"]
    )
    m.fs.MX4 = Mixer(
        property_package=m.fs.props_ASM1, inlet_list=["thickener", "clarifier"]
    )

    # Make connections related to AD section
    m.fs.stream2adm = Arc(
        source=m.fs.RADM.liquid_outlet, destination=m.fs.adm_asm.inlet
    )
    m.fs.stream6adm = Arc(source=m.fs.SP6.waste, destination=m.fs.TU.inlet)
    m.fs.stream3adm = Arc(source=m.fs.TU.underflow, destination=m.fs.MX4.thickener)
    m.fs.stream7adm = Arc(source=m.fs.TU.overflow, destination=m.fs.MX3.recycle2)
    m.fs.stream9adm = Arc(source=m.fs.CL.underflow, destination=m.fs.MX4.clarifier)
    m.fs.stream4adm = Arc(source=m.fs.adm_asm.outlet, destination=m.fs.DU.inlet)
    m.fs.stream5adm = Arc(source=m.fs.DU.overflow, destination=m.fs.MX2.recycle1)
    m.fs.stream01 = Arc(source=m.fs.FeedWater.outlet, destination=m.fs.MX2.feed_water1)
    m.fs.stream02 = Arc(source=m.fs.MX2.outlet, destination=m.fs.MX3.feed_water2)
    m.fs.stream03 = Arc(source=m.fs.MX3.outlet, destination=m.fs.CL.inlet)
    m.fs.stream04 = Arc(source=m.fs.CL.effluent, destination=m.fs.MX1.feed_water)
    m.fs.stream10adm = Arc(source=m.fs.MX4.outlet, destination=m.fs.asm_adm.inlet)
    m.fs.stream1adm = Arc(source=m.fs.asm_adm.outlet, destination=m.fs.RADM.inlet)

    pyo.TransformationFactory("network.expand_arcs").apply_to(m)

    iscale.calculate_scaling_factors(m.fs)

    return m


def set_operating_conditions(m):
    # Feed Water Conditions
    m.fs.FeedWater.flow_vol.fix(20648 * pyo.units.m**3 / pyo.units.day)
    m.fs.FeedWater.temperature.fix(308.15 * pyo.units.K)
    m.fs.FeedWater.pressure.fix(1 * pyo.units.atm)
    m.fs.FeedWater.conc_mass_comp[0, "S_I"].fix(27 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "S_S"].fix(58 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "X_I"].fix(92 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "X_S"].fix(363 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "X_BH"].fix(50 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "X_BA"].fix(0 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "X_P"].fix(0 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "S_O"].fix(0 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "S_NO"].fix(0 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "S_NH"].fix(23 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "S_ND"].fix(5 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.conc_mass_comp[0, "X_ND"].fix(16 * pyo.units.g / pyo.units.m**3)
    m.fs.FeedWater.alkalinity.fix(7 * pyo.units.mol / pyo.units.m**3)

    # Reactor sizing in activated sludge process
    m.fs.R1.volume.fix(1000 * pyo.units.m**3)
    m.fs.R2.volume.fix(1000 * pyo.units.m**3)
    m.fs.R3.volume.fix(1333 * pyo.units.m**3)
    m.fs.R4.volume.fix(1333 * pyo.units.m**3)
    m.fs.R5.volume.fix(1333 * pyo.units.m**3)

    # Injection rates to Reactors 3, 4 and 5 of the activated sludge process
    for j in m.fs.props_ASM1.component_list:
        if j != "S_O":
            # All components except S_O have no injection
            m.fs.R3.injection[:, :, j].fix(0)
            m.fs.R4.injection[:, :, j].fix(0)
            m.fs.R5.injection[:, :, j].fix(0)
    # Then set injections rates for O2
    m.fs.R3.outlet.conc_mass_comp[:, "S_O"].fix(1.72e-3)
    m.fs.R4.outlet.conc_mass_comp[:, "S_O"].fix(2.43e-3)
    m.fs.R5.outlet.conc_mass_comp[:, "S_O"].fix(4.49e-4)

    # Set fraction of outflow from reactor 5 that goes to recycle
    m.fs.SP5.split_fraction[:, "underflow"].fix(0.6)

    # Secondary clarifier
    # TODO: Update once secondary clarifier with more detailed model available
    m.fs.CL1.split_fraction[0, "effluent", "H2O"].fix(0.48956)
    m.fs.CL1.split_fraction[0, "effluent", "S_I"].fix(0.48956)
    m.fs.CL1.split_fraction[0, "effluent", "S_S"].fix(0.48956)
    m.fs.CL1.split_fraction[0, "effluent", "X_I"].fix(0.00187)
    m.fs.CL1.split_fraction[0, "effluent", "X_S"].fix(0.00187)
    m.fs.CL1.split_fraction[0, "effluent", "X_BH"].fix(0.00187)
    m.fs.CL1.split_fraction[0, "effluent", "X_BA"].fix(0.00187)
    m.fs.CL1.split_fraction[0, "effluent", "X_P"].fix(0.00187)
    m.fs.CL1.split_fraction[0, "effluent", "S_O"].fix(0.48956)
    m.fs.CL1.split_fraction[0, "effluent", "S_NO"].fix(0.48956)
    m.fs.CL1.split_fraction[0, "effluent", "S_NH"].fix(0.48956)
    m.fs.CL1.split_fraction[0, "effluent", "S_ND"].fix(0.48956)
    m.fs.CL1.split_fraction[0, "effluent", "X_ND"].fix(0.00187)
    m.fs.CL1.split_fraction[0, "effluent", "S_ALK"].fix(0.48956)

    m.fs.CL1.surface_area.fix(1500 * pyo.units.m**2)

    # Sludge purge separator
    m.fs.SP6.split_fraction[:, "recycle"].fix(0.985)

    # Outlet pressure from recycle pump
    m.fs.P1.outlet.pressure.fix(101325)

    # Primary Clarifier
    # TODO: Update primary clarifier once more detailed model available
    m.fs.CL.split_fraction[0, "effluent", "H2O"].fix(0.993)
    m.fs.CL.split_fraction[0, "effluent", "S_I"].fix(0.993)
    m.fs.CL.split_fraction[0, "effluent", "S_S"].fix(0.993)
    m.fs.CL.split_fraction[0, "effluent", "X_I"].fix(0.5192)
    m.fs.CL.split_fraction[0, "effluent", "X_S"].fix(0.5192)
    m.fs.CL.split_fraction[0, "effluent", "X_BH"].fix(0.5192)
    m.fs.CL.split_fraction[0, "effluent", "X_BA"].fix(0.5192)
    m.fs.CL.split_fraction[0, "effluent", "X_P"].fix(0.5192)
    m.fs.CL.split_fraction[0, "effluent", "S_O"].fix(0.993)
    m.fs.CL.split_fraction[0, "effluent", "S_NO"].fix(0.993)
    m.fs.CL.split_fraction[0, "effluent", "S_NH"].fix(0.993)
    m.fs.CL.split_fraction[0, "effluent", "S_ND"].fix(0.993)
    m.fs.CL.split_fraction[0, "effluent", "X_ND"].fix(0.5192)
    m.fs.CL.split_fraction[0, "effluent", "S_ALK"].fix(0.993)

    # Anaerobic digester
    m.fs.RADM.volume_liquid.fix(3400)
    m.fs.RADM.volume_vapor.fix(300)
    m.fs.RADM.liquid_outlet.temperature.fix(308.15)

    # Dewatering Unit - fix either HRT or volume.
    m.fs.DU.hydraulic_retention_time.fix(1800 * pyo.units.s)

    # Set specific energy consumption averaged for centrifuge
    m.fs.DU.energy_electric_flow_vol_inlet[0] = 0.069 * pyo.units.kWh / pyo.units.m**3

    # Thickener unit
    m.fs.TU.hydraulic_retention_time.fix(86400 * pyo.units.s)
    m.fs.TU.diameter.fix(10 * pyo.units.m)


def initialize_system(m):
    # Initialize flowsheet
    # Apply sequential decomposition - 1 iteration should suffice
    seq = SequentialDecomposition()
    seq.options.tear_method = "Direct"
    seq.options.iterLim = 1
    seq.options.tear_set = [m.fs.stream2, m.fs.stream10adm]

    G = seq.create_graph(m)
    # Uncomment this code to see tear set and initialization order
    order = seq.calculation_order(G)
    print("Initialization Order")
    for o in order:
        print(o[0].name)

    # Initial guesses for flow into first reactor
    tear_guesses1 = {
        "flow_vol": {0: 103531 / 24 / 3600},
        "conc_mass_comp": {
            (0, "S_I"): 0.028,
            (0, "S_S"): 0.012,
            (0, "X_I"): 1.532,
            (0, "X_S"): 0.069,
            (0, "X_BH"): 2.233,
            (0, "X_BA"): 0.167,
            (0, "X_P"): 0.964,
            (0, "S_O"): 0.0011,
            (0, "S_NO"): 0.0073,
            (0, "S_NH"): 0.0072,
            (0, "S_ND"): 0.0016,
            (0, "X_ND"): 0.0040,
        },
        "alkalinity": {0: 0.0052},
        "temperature": {0: 308.15},
        "pressure": {0: 101325},
    }

    tear_guesses2 = {
        "flow_vol": {0: 178 / 24 / 3600},
        "conc_mass_comp": {
            (0, "S_I"): 0.028,
            (0, "S_S"): 0.048,
            (0, "X_I"): 10.362,
            (0, "X_S"): 20.375,
            (0, "X_BH"): 10.210,
            (0, "X_BA"): 0.553,
            (0, "X_P"): 3.204,
            (0, "S_O"): 0.00025,
            (0, "S_NO"): 0.00169,
            (0, "S_NH"): 0.0289,
            (0, "S_ND"): 0.00468,
            (0, "X_ND"): 0.906,
        },
        "alkalinity": {0: 0.00715},
        "temperature": {0: 308.15},
        "pressure": {0: 101325},
    }

    # Pass the tear_guess to the SD tool
    seq.set_guesses_for(m.fs.R1.inlet, tear_guesses1)
    seq.set_guesses_for(m.fs.asm_adm.inlet, tear_guesses2)

    def function(unit):
        unit.initialize(outlvl=idaeslog.INFO_HIGH)

    seq.run(m, function)


def add_costing(m):
    m.fs.costing = WaterTAPCosting()
    m.fs.costing.base_currency = pyo.units.USD_2020

    # Costing Blocks
    m.fs.R1.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)
    m.fs.R2.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)
    m.fs.R3.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)
    m.fs.R4.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)
    m.fs.R5.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)
    m.fs.CL.costing = UnitModelCostingBlock(
        flowsheet_costing_block=m.fs.costing,
        costing_method=cost_primary_clarifier,
    )

    m.fs.CL1.costing = UnitModelCostingBlock(
        flowsheet_costing_block=m.fs.costing,
        costing_method=cost_circular_clarifier,
    )

    m.fs.RADM.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)
    m.fs.DU.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)
    m.fs.TU.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)

    # Leaving out mixer costs for now
    # m.fs.MX1.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)
    # m.fs.MX6.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)
    # m.fs.MX2.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)
    # m.fs.MX3.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)
    # m.fs.MX4.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)

    # process costing and add system level metrics
    m.fs.costing.cost_process()
    m.fs.costing.add_electricity_intensity(m.fs.FeedWater.properties[0].flow_vol)
    m.fs.costing.add_annual_water_production(m.fs.Treated.properties[0].flow_vol)
    m.fs.costing.add_LCOW(m.fs.FeedWater.properties[0].flow_vol)
    m.fs.costing.add_specific_energy_consumption(m.fs.FeedWater.properties[0].flow_vol)

    m.fs.objective = pyo.Objective(expr=m.fs.costing.LCOW)
    iscale.calculate_scaling_factors(m.fs)


def solve(blk, solver=None):
    if solver is None:
        solver = get_solver()
    results = solver.solve(blk, tee=True)
    pyo.assert_optimal_termination(results)
    return results


def display_results(m):
    m.display()

    unit_list = [
        "FeedWater",
        "MX1",
        "R1",
        "R2",
        "R3",
        "R4",
        "R5",
        "SP5",
        "CL1",
        "SP6",
        "MX6",
        "Treated",
        "P1",
        "asm_adm",
        "RADM",
        "adm_asm",
        "CL",
        "TU",
        "DU",
        "MX2",
        "MX3",
        "MX4",
    ]
    for u in unit_list:
        m.fs.component(u).report()


def display_costing(m):
    print("Levelized cost of treatment: %.2f $/m3" % pyo.value(m.fs.costing.LCOW))

    print(
        "Total operating cost: %.2f $/yr" % pyo.value(m.fs.costing.total_operating_cost)
    )
    print("Total capital cost: %.2f $" % pyo.value(m.fs.costing.total_capital_cost))

    print(
        "Total annualized cost: %.2f $/yr"
        % pyo.value(m.fs.costing.total_annualized_cost)
    )
    print(
        "Specific energy consumption with respect to influent flowrate: %.1f kWh/m3"
        % pyo.value(m.fs.costing.specific_energy_consumption)
    )

    print(
        "electricity consumption R3",
        pyo.value(m.fs.R3.electricity_consumption[0]),
        pyo.units.get_units(m.fs.R3.electricity_consumption[0]),
    )
    print(
        "electricity consumption R4",
        pyo.value(m.fs.R4.electricity_consumption[0]),
        pyo.units.get_units(m.fs.R4.electricity_consumption[0]),
    )
    print(
        "electricity consumption R5",
        pyo.value(m.fs.R5.electricity_consumption[0]),
        pyo.units.get_units(m.fs.R5.electricity_consumption[0]),
    )
    print(
        "electricity consumption primary clarifier",
        pyo.value(m.fs.CL.electricity_consumption[0]),
        pyo.units.get_units(m.fs.CL.electricity_consumption[0]),
    )
    print(
        "electricity consumption secondary clarifier",
        pyo.value(m.fs.CL1.electricity_consumption[0]),
        pyo.units.get_units(m.fs.CL1.electricity_consumption[0]),
    )
    print(
        "electricity consumption AD",
        pyo.value(m.fs.RADM.electricity_consumption[0]),
        pyo.units.get_units(m.fs.RADM.electricity_consumption[0]),
    )
    print(
        "electricity consumption dewatering Unit",
        pyo.value(m.fs.DU.electricity_consumption[0]),
        pyo.units.get_units(m.fs.DU.electricity_consumption[0]),
    )
    print(
        "electricity consumption thickening Unit",
        pyo.value(m.fs.TU.electricity_consumption[0]),
        pyo.units.get_units(m.fs.TU.electricity_consumption[0]),
    )
    print(
        "Influent flow",
        pyo.value(m.fs.FeedWater.flow_vol[0]),
        pyo.units.get_units(m.fs.FeedWater.flow_vol[0]),
    )
    print(
        "flow into R3",
        pyo.value(m.fs.R3.control_volume.properties_in[0].flow_vol),
        pyo.units.get_units(m.fs.R3.control_volume.properties_in[0].flow_vol),
    )
    print(
        "flow into RADM",
        pyo.value(m.fs.RADM.liquid_phase.properties_in[0].flow_vol),
        pyo.units.get_units(m.fs.RADM.liquid_phase.properties_in[0].flow_vol),
    )

    print(
        "capital cost R1",
        pyo.value(m.fs.R1.costing.capital_cost),
        pyo.units.get_units(m.fs.R1.costing.capital_cost),
    )
    print(
        "capital cost R2",
        pyo.value(m.fs.R2.costing.capital_cost),
        pyo.units.get_units(m.fs.R2.costing.capital_cost),
    )
    print(
        "capital cost R3",
        pyo.value(m.fs.R3.costing.capital_cost),
        pyo.units.get_units(m.fs.R3.costing.capital_cost),
    )
    print(
        "capital cost R4",
        pyo.value(m.fs.R4.costing.capital_cost),
        pyo.units.get_units(m.fs.R4.costing.capital_cost),
    )
    print(
        "capital cost R5",
        pyo.value(m.fs.R5.costing.capital_cost),
        pyo.units.get_units(m.fs.R5.costing.capital_cost),
    )
    print(
        "capital cost primary clarifier",
        pyo.value(m.fs.CL.costing.capital_cost),
        pyo.units.get_units(m.fs.CL.costing.capital_cost),
    )
    print(
        "capital cost secondary clarifier",
        pyo.value(m.fs.CL1.costing.capital_cost),
        pyo.units.get_units(m.fs.CL1.costing.capital_cost),
    )
    print(
        "capital cost AD",
        pyo.value(m.fs.RADM.costing.capital_cost),
        pyo.units.get_units(m.fs.RADM.costing.capital_cost),
    )
    print(
        "capital cost dewatering Unit",
        pyo.value(m.fs.DU.costing.capital_cost),
        pyo.units.get_units(m.fs.DU.costing.capital_cost),
    )
    print(
        "capital cost thickener unit",
        pyo.value(m.fs.TU.costing.capital_cost),
        pyo.units.get_units(m.fs.TU.costing.capital_cost),
    )


if __name__ == "__main__":
    m, results = main()
